import numpy as np
from numpy.testing import assert_, run_module_suite
import scipy.linalg as la

from qutip import (spre, spost, qeye, sigmax, sigmay, sigmaz, qpt)
from qutip.qip.operations.gates import snot, cnot


def test_qpt_snot():
    "quantum process tomography for snot gate"

    U_psi = snot()
    U_rho = spre(U_psi) * spost(U_psi.dag())
    N = 1
    op_basis = [[qeye(2), sigmax(), 1j * sigmay(), sigmaz()] for i in range(N)]
    # op_label = [["i", "x", "y", "z"] for i in range(N)]
    chi1 = qpt(U_rho, op_basis)

    chi2 = np.zeros((2 ** (2 * N), 2 ** (2 * N)), dtype=complex)
    chi2[1, 1] = chi2[1, 3] = chi2[3, 1] = chi2[3, 3] = 0.5

    assert_(la.norm(chi2 - chi1) < 1e-8)


def test_qpt_cnot():
    "quantum process tomography for cnot gate"

    U_psi = cnot()
    U_rho = spre(U_psi) * spost(U_psi.dag())
    N = 2
    op_basis = [[qeye(2), sigmax(), 1j * sigmay(), sigmaz()] for i in range(N)]
    # op_label = [["i", "x", "y", "z"] for i in range(N)]
    chi1 = qpt(U_rho, op_basis)

    chi2 = np.zeros((2 ** (2 * N), 2 ** (2 * N)), dtype=complex)
    chi2[0, 0] = chi2[0, 1] = chi2[1, 0] = chi2[1, 1] = 0.25

    chi2[12, 0] = chi2[12, 1] = 0.25
    chi2[13, 0] = chi2[13, 1] = -0.25

    chi2[0, 12] = chi2[1, 12] = 0.25
    chi2[0, 13] = chi2[1, 13] = -0.25

    chi2[12, 12] = chi2[13, 13] = 0.25
    chi2[13, 12] = chi2[12, 13] = -0.25

    assert_(la.norm(chi2 - chi1) < 1e-8)

if __name__ == "__main__":
    run_module_suite()



import pytest
import numpy as np
import uuid
import qutip
from pathlib import Path

# qsave _always_ appends a suffix to the file name at the time of writing, but
# in case this changes in the future, to ensure that we never leak a temporary
# file into the user's folders, we simply apply these tests in a temporary
# directory.  Windows also does not allow temporary files to be opened multiple
# times, so using a temporary directory is best.
pytestmark = [pytest.mark.usefixtures("in_temporary_directory")]

_dimension = 10

def _random_file_name():
    return "_" + str(uuid.uuid4())


class Test_file_data_store_file_data_read:
    # Tests parametrised seprately to give nicer descriptions in verbose mode.

    def case(self, filename, kwargs):
        data = 1 - 2*np.random.rand(_dimension, _dimension)
        if kwargs.get('numtype', 'complex') == 'complex':
            data = data * (0.5*0.5j)
        qutip.file_data_store(filename, data, **kwargs)
        out = qutip.file_data_read(filename)
        np.testing.assert_allclose(data, out, atol=1e-8)

    def test_defaults(self):
        return self.case(_random_file_name(), {})

    @pytest.mark.parametrize("type_", ["real", "complex"])
    @pytest.mark.parametrize("format_", ["decimal", "exp"])
    def test_type_format(self, type_, format_):

        kwargs = {'numtype': type_, 'numformat': format_}
        return self.case(_random_file_name(), kwargs)

    @pytest.mark.parametrize("separator", [",", ";", "\t", " ", " \t "],
                             ids=lambda x: "'" + x + "'")
    def test_separator_detection(self, separator):
        kwargs = {'numtype': 'complex', 'numformat': 'exp', 'sep': separator}
        return self.case(_random_file_name(), kwargs)


@pytest.mark.parametrize('use_path', [True, False], ids=['Path', 'str'])
@pytest.mark.parametrize('suffix', ['', '.qu', '.dat'])
def test_qsave_qload(use_path, suffix):
    ops_in = [qutip.sigmax(),
              qutip.num(_dimension),
              qutip.coherent_dm(_dimension, 1j)]
    filename = _random_file_name() + suffix
    if use_path:
        filename = Path.cwd() / filename
    qutip.qsave(ops_in, filename)
    ops_out = qutip.qload(filename)
    assert ops_in == ops_out



#cython: language_level=3

import numpy as np
from qutip.fastsparse import fast_csr_matrix
cimport numpy as cnp
from libc.math cimport abs, fabs, sqrt
from libcpp cimport bool
cimport cython
cnp.import_array()

cdef extern from "numpy/arrayobject.h" nogil:
    void PyArray_ENABLEFLAGS(cnp.ndarray arr, int flags)
    void PyDataMem_FREE(void * ptr)
    void PyDataMem_RENEW(void * ptr, size_t size)
    void PyDataMem_NEW_ZEROED(size_t size, size_t elsize)
    void PyDataMem_NEW(size_t size)


cdef extern from "<complex>" namespace "std" nogil:
    double abs(double complex x)
    double real(double complex x)
    double imag(double complex x)

cdef extern from "<complex>" namespace "std" nogil:
    double cabs "abs" (double complex x)

cdef inline int int_max(int x, int y):
    return x ^ ((x ^ y) & -(x < y))

include "parameters.pxi"

@cython.boundscheck(False)
@cython.wraparound(False)
def _sparse_bandwidth(
        int[::1] idx,
        int[::1] ptr,
        int nrows):
    """
    Calculates the max (mb), lower(lb), and upper(ub) bandwidths of a
    csr_matrix.
    """
    cdef int ldist
    cdef int lb = -nrows
    cdef int ub = -nrows
    cdef int mb = 0
    cdef size_t ii, jj

    for ii in range(nrows):
        for jj in range(ptr[ii], ptr[ii + 1]):
            ldist = ii - idx[jj]
            lb = int_max(lb, ldist)
            ub = int_max(ub, -ldist)
            mb = int_max(mb, ub + lb + 1)

    return mb, lb, ub


@cython.boundscheck(False)
@cython.wraparound(False)
def _sparse_profile(int[::1] idx,
        int[::1] ptr,
        int nrows):
    cdef int ii, jj, temp, ldist=0
    cdef LTYPE_t pro = 0
    for ii in range(nrows):
        temp = 0
        for jj in range(ptr[ii], ptr[ii + 1]):
            ldist = idx[jj] - ii
            temp = int_max(temp, ldist)
        pro += temp
    return pro


@cython.boundscheck(False)
@cython.wraparound(False)
def _sparse_permute(
        cnp.ndarray[cython.numeric, ndim=1] data,
        int[::1] idx,
        int[::1] ptr,
        int nrows,
        int ncols,
        cnp.ndarray[ITYPE_t, ndim=1] rperm,
        cnp.ndarray[ITYPE_t, ndim=1] cperm,
        int flag):
    """
    Permutes the rows and columns of a sparse CSR or CSC matrix according to
    the permutation arrays rperm and cperm, respectively.
    Here, the permutation arrays specify the new order of the rows and columns.
    i.e. [0,1,2,3,4] -> [3,0,4,1,2].
    """
    cdef int ii, jj, kk, k0, nnz
    cdef cnp.ndarray[cython.numeric] new_data = np.zeros_like(data)
    cdef cnp.ndarray[ITYPE_t] new_idx = np.zeros_like(idx)
    cdef cnp.ndarray[ITYPE_t] new_ptr = np.zeros_like(ptr)
    cdef cnp.ndarray[ITYPE_t] perm_r
    cdef cnp.ndarray[ITYPE_t] perm_c
    cdef cnp.ndarray[ITYPE_t] inds

    if flag == 0:  # CSR matrix
        if rperm.shape[0] != 0:
            inds = np.argsort(rperm).astype(ITYPE)
            perm_r = np.arange(rperm.shape[0], dtype=ITYPE)[inds]

            for jj in range(nrows):
                ii = perm_r[jj]
                new_ptr[ii + 1] = ptr[jj + 1] - ptr[jj]

            for jj in range(nrows):
                new_ptr[jj + 1] = new_ptr[jj+1] + new_ptr[jj]

            for jj in range(nrows):
                k0 = new_ptr[perm_r[jj]]
                for kk in range(ptr[jj], ptr[jj + 1]):
                    new_idx[k0] = idx[kk]
                    new_data[k0] = data[kk]
                    k0 = k0 + 1

        if cperm.shape[0] != 0:
            inds = np.argsort(cperm).astype(ITYPE)
            perm_c = np.arange(cperm.shape[0], dtype=ITYPE)[inds]
            nnz = new_ptr[new_ptr.shape[0] - 1]
            for jj in range(nnz):
                new_idx[jj] = perm_c[new_idx[jj]]

    elif flag == 1:  # CSC matrix
        if cperm.shape[0] != 0:
            inds = np.argsort(cperm).astype(ITYPE)
            perm_c = np.arange(cperm.shape[0], dtype=ITYPE)[inds]

            for jj in range(ncols):
                ii = perm_c[jj]
                new_ptr[ii + 1] = ptr[jj + 1] - ptr[jj]

            for jj in range(ncols):
                new_ptr[jj + 1] = new_ptr[jj + 1] + new_ptr[jj]

            for jj in range(ncols):
                k0 = new_ptr[perm_c[jj]]
                for kk in range(ptr[jj], ptr[jj + 1]):
                    new_idx[k0] = idx[kk]
                    new_data[k0] = data[kk]
                    k0 = k0 + 1

        if rperm.shape[0] != 0:
            inds = np.argsort(rperm).astype(ITYPE)
            perm_r = np.arange(rperm.shape[0], dtype=ITYPE)[inds]
            nnz = new_ptr[new_ptr.shape[0] - 1]
            for jj in range(nnz):
                new_idx[jj] = perm_r[new_idx[jj]]

    return new_data, new_idx, new_ptr


@cython.boundscheck(False)
@cython.wraparound(False)
def _sparse_reverse_permute(
        cnp.ndarray[cython.numeric, ndim=1] data,
        int[::1] idx,
        int[::1] ptr,
        int nrows,
        int ncols,
        cnp.ndarray[ITYPE_t, ndim=1] rperm,
        cnp.ndarray[ITYPE_t, ndim=1] cperm,
        int flag):
    """
    Reverse permutes the rows and columns of a sparse CSR or CSC matrix
    according to the original permutation arrays rperm and cperm, respectively.
    """
    cdef int ii, jj, kk, k0, nnz
    cdef cnp.ndarray[cython.numeric, ndim=1] new_data = np.zeros_like(data)
    cdef cnp.ndarray[ITYPE_t, ndim=1] new_idx = np.zeros_like(idx)
    cdef cnp.ndarray[ITYPE_t, ndim=1] new_ptr = np.zeros_like(ptr)

    if flag == 0:  # CSR matrix
        if rperm.shape[0] != 0:
            for jj in range(nrows):
                ii = rperm[jj]
                new_ptr[ii + 1] = ptr[jj + 1] - ptr[jj]

            for jj in range(nrows):
                new_ptr[jj + 1] = new_ptr[jj + 1] + new_ptr[jj]

            for jj in range(nrows):
                k0 = new_ptr[rperm[jj]]
                for kk in range(ptr[jj], ptr[jj + 1]):
                    new_idx[k0] = idx[kk]
                    new_data[k0] = data[kk]
                    k0 = k0 + 1

        if cperm.shape[0] > 0:
            nnz = new_ptr[new_ptr.shape[0] - 1]
            for jj in range(nnz):
                new_idx[jj] = cperm[new_idx[jj]]

    if flag == 1:  # CSC matrix
        if cperm.shape[0] != 0:
            for jj in range(ncols):
                ii = cperm[jj]
                new_ptr[ii + 1] = ptr[jj + 1] - ptr[jj]

            for jj in range(ncols):
                new_ptr[jj + 1] = new_ptr[jj + 1] + new_ptr[jj]

            for jj in range(ncols):
                k0 = new_ptr[cperm[jj]]
                for kk in range(ptr[jj], ptr[jj + 1]):
                    new_idx[k0] = idx[kk]
                    new_data[k0] = data[kk]
                    k0 = k0 + 1

        if cperm.shape[0] != 0:
            nnz = new_ptr[new_ptr.shape[0] - 1]
            for jj in range(nnz):
                new_idx[jj] = rperm[new_idx[jj]]

    return new_data, new_idx, new_ptr


@cython.boundscheck(False)
@cython.wraparound(False)
def _isdiag(int[::1] idx,
        int[::1] ptr,
        int nrows):

    cdef int row, num_elems
    for row in range(nrows):
        num_elems = ptr[row+1] - ptr[row]
        if num_elems > 1:
            return 0
        elif num_elems == 1:
            if idx[ptr[row]] != row:
                return 0
    return 1


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef cnp.ndarray[complex, ndim=1, mode='c'] _csr_get_diag(complex[::1] data,
    int[::1] idx, int[::1] ptr, int k=0):

    cdef size_t row, jj
    cdef int num_rows = ptr.shape[0]-1
    cdef int abs_k = abs(k)
    cdef int start, stop
    cdef cnp.ndarray[complex, ndim=1, mode='c'] out = np.zeros(num_rows-abs_k, dtype=complex)

    if k >= 0:
        start = 0
        stop = num_rows-abs_k
    else: #k < 0
        start = abs_k
        stop = num_rows

    for row in range(start, stop):
        for jj in range(ptr[row], ptr[row+1]):
            if idx[jj]-k == row:
                out[row-start] = data[jj]
                break
    return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def unit_row_norm(complex[::1] data, int[::1] ptr, int nrows):
    cdef size_t row, ii
    cdef double total
    for row in range(nrows):
        total = 0
        for ii in range(ptr[row], ptr[row+1]):
            total += real(data[ii]) * real(data[ii]) + imag(data[ii]) * imag(data[ii])
        total = sqrt(total)
        for ii in range(ptr[row], ptr[row+1]):
            data[ii] /= total



@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double zcsr_one_norm(complex[::1] data, int[::1] ind, int[::1] ptr,
                     int nrows, int ncols):

    cdef int k
    cdef size_t ii, jj
    cdef double * col_sum = <double *>PyDataMem_NEW_ZEROED(ncols, sizeof(double))
    cdef double max_col = 0
    for ii in range(nrows):
        for jj in range(ptr[ii], ptr[ii+1]):
            k = ind[jj]
            col_sum[k] += cabs(data[jj])
    for ii in range(ncols):
        if col_sum[ii] > max_col:
            max_col = col_sum[ii]
    PyDataMem_FREE(col_sum)
    return max_col


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double zcsr_inf_norm(complex[::1] data, int[::1] ind, int[::1] ptr,
                     int nrows, int ncols):

    cdef int k
    cdef size_t ii, jj
    cdef double * row_sum = <double *>PyDataMem_NEW_ZEROED(nrows, sizeof(double))
    cdef double max_row = 0
    for ii in range(nrows):
        for jj in range(ptr[ii], ptr[ii+1]):
            row_sum[ii] += cabs(data[jj])
    for ii in range(nrows):
        if row_sum[ii] > max_row:
            max_row = row_sum[ii]
    PyDataMem_FREE(row_sum)
    return max_row


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef bool cy_tidyup(complex[::1] data, double atol, unsigned int nnz):
    """
    Performs an in-place tidyup of CSR matrix data
    """
    cdef size_t kk
    cdef double re, im
    cdef bool re_flag, im_flag, out_flag = 0
    for kk in range(nnz):
        re_flag = 0
        im_flag = 0
        re = real(data[kk])
        im = imag(data[kk])
        if fabs(re) < atol:
            re = 0
            re_flag = 1
        if fabs(im) < atol:
            im = 0
            im_flag = 1

        if re_flag or im_flag:
            data[kk] = re + 1j*im

        if re_flag and im_flag:
            out_flag = 1
    return out_flag



from qutip.qip.device.processor import Processor
from qutip.qip.device.spinchain import (
    SpinChain, LinearSpinChain, CircularSpinChain)
from qutip.qip.device.cavityqed import DispersiveCavityQED
from qutip.qip.device.optpulseprocessor import OptPulseProcessor
from qutip.qip.device.modelprocessor import ModelProcessor



#cython: language_level=3

cimport cython
from qutip.cy.spmatfuncs cimport spmvpy
from qutip.cy.openmp.parfuncs cimport spmvpy_openmp


@cython.boundscheck(False)
@cython.wraparound(False)
def _spmvpy(complex[::1] data,
            int[::1] ind,
            int[::1] ptr,
            complex[::1] vec,
            complex a,
            complex[::1] out):
    spmvpy(&data[0], &ind[0], &ptr[0], &vec[0], a, &out[0], vec.shape[0])



@cython.boundscheck(False)
@cython.wraparound(False)
def _spmvpy_openmp(complex[::1] data,
            int[::1] ind,
            int[::1] ptr,
            complex[::1] vec,
            complex a,
            complex[::1] out,
            unsigned int num_threads):
    spmvpy_openmp(&data[0], &ind[0], &ptr[0], &vec[0], a, &out[0], vec.shape[0], num_threads)



#cython: language_level=3

from qutip.cy.sparse_structs cimport CSR_Matrix

cdef void fdense2D_to_CSR(complex[::1, :] mat, CSR_Matrix * out,
                                unsigned int nrows, unsigned int ncols)



#cython: language_level=3

cimport numpy as cnp
cimport cython

cpdef cnp.ndarray[complex, ndim=1, mode="c"] spmv_csr_openmp(complex[::1] data,
                int[::1] ind, int[::1] ptr, complex[::1] vec, unsigned int nthr)


cdef void spmvpy_openmp(complex * data,
                int * ind,
                int *  ptr,
                complex * vec,
                complex a,
                complex * out,
                unsigned int nrows,
                unsigned int nthr)



__all__ = ['floquet_modes', 'floquet_modes_t', 'floquet_modes_table',
           'floquet_modes_t_lookup', 'floquet_states', 'floquet_states_t',
           'floquet_wavefunction', 'floquet_wavefunction_t',
           'floquet_state_decomposition', 'fsesolve',
           'floquet_master_equation_rates', 'floquet_collapse_operators',
           'floquet_master_equation_tensor',
           'floquet_master_equation_steadystate', 'floquet_basis_transform',
           'floquet_markov_mesolve', 'fmmesolve']

import numpy as np
import scipy.linalg as la
import scipy
import warnings
from copy import copy
from numpy import angle, pi, exp, sqrt
from types import FunctionType
from qutip.qobj import Qobj, isket
from qutip.superoperator import vec2mat_index, mat2vec, vec2mat
#from qutip.mesolve import mesolve
from qutip.sesolve import sesolve
from qutip.rhs_generate import rhs_clear
from qutip.steadystate import steadystate
from qutip.states import ket2dm
from qutip.states import projection
from qutip.solver import Options
from qutip.propagator import propagator
from qutip.solver import Result, _solver_safety_check
from qutip.cy.spmatfuncs import cy_ode_rhs
from qutip.expect import expect
from qutip.utilities import n_thermal


def floquet_modes(H, T, args=None, sort=False, U=None, options=None):
    """
    Calculate the initial Floquet modes Phi_alpha(0) for a driven system with
    period T.

    Returns a list of :class:`qutip.qobj` instances representing the Floquet
    modes and a list of corresponding quasienergies, sorted by increasing
    quasienergy in the interval [-pi/T, pi/T]. The optional parameter `sort`
    decides if the output is to be sorted in increasing quasienergies or not.

    Parameters
    ----------

    H : :class:`qutip.qobj`
        system Hamiltonian, time-dependent with period `T`

    args : dictionary
        dictionary with variables required to evaluate H

    T : float
        The period of the time-dependence of the hamiltonian. The default value
        'None' indicates that the 'tlist' spans a single period of the driving.

    U : :class:`qutip.qobj`
        The propagator for the time-dependent Hamiltonian with period `T`.
        If U is `None` (default), it will be calculated from the Hamiltonian
        `H` using :func:`qutip.propagator.propagator`.

    options : :class:`qutip.solver`
        options for the ODE solver. For the propagator U.

    Returns
    -------

    output : list of kets, list of quasi energies

        Two lists: the Floquet modes as kets and the quasi energies.

    """

    if U is None:
        # get the unitary propagator
        U = propagator(H, T, [], args=args, options=copy(options))

    # find the eigenstates for the propagator
    evals, evecs = la.eig(U.full())

    eargs = angle(evals)

    # make sure that the phase is in the interval [-pi, pi], so that
    # the quasi energy is in the interval [-pi/T, pi/T] where T is the
    # period of the driving.  eargs += (eargs <= -2*pi) * (2*pi) +
    # (eargs > 0) * (-2*pi)
    eargs += (eargs <= -pi) * (2 * pi) + (eargs > pi) * (-2 * pi)
    e_quasi = -eargs / T

    # sort by the quasi energy
    if sort:
        order = np.argsort(-e_quasi)
    else:
        order = list(range(len(evals)))

    # prepare a list of kets for the floquet states
    new_dims = [U.dims[0], [1] * len(U.dims[0])]
    new_shape = [U.shape[0], 1]
    kets_order = [Qobj(np.array(evecs[:, o]).T,
                       dims=new_dims, shape=new_shape) for o in order]

    return kets_order, e_quasi[order]


def floquet_modes_t(f_modes_0, f_energies, t, H, T, args=None,
                    options=None):
    """
    Calculate the Floquet modes at times tlist Phi_alpha(tlist) propagting the
    initial Floquet modes Phi_alpha(0)

    Parameters
    ----------

    f_modes_0 : list of :class:`qutip.qobj` (kets)
        Floquet modes at :math:`t`

    f_energies : list
        Floquet energies.

    t : float
        The time at which to evaluate the floquet modes.

    H : :class:`qutip.qobj`
        system Hamiltonian, time-dependent with period `T`

    args : dictionary
        dictionary with variables required to evaluate H

    T : float
        The period of the time-dependence of the hamiltonian.

    options : :class:`qutip.solver`
        options for the ODE solver. For the propagator.

    Returns
    -------

    output : list of kets

        The Floquet modes as kets at time :math:`t`

    """
    # find t in [0,T] such that t_orig = t + n * T for integer n
    t = t - int(t / T) * T
    f_modes_t = []

    # get the unitary propagator from 0 to t
    if t > 0.0:
        U = propagator(H, t, [], args, options=copy(options))

        for n in np.arange(len(f_modes_0)):
            f_modes_t.append(U * f_modes_0[n] * exp(1j * f_energies[n] * t))
    else:
        f_modes_t = f_modes_0

    return f_modes_t


def floquet_modes_table(f_modes_0, f_energies, tlist, H, T, args=None,
                        options=None):
    """
    Pre-calculate the Floquet modes for a range of times spanning the floquet
    period. Can later be used as a table to look up the floquet modes for
    any time.

    Parameters
    ----------

    f_modes_0 : list of :class:`qutip.qobj` (kets)
        Floquet modes at :math:`t`

    f_energies : list
        Floquet energies.

    tlist : array
        The list of times at which to evaluate the floquet modes.

    H : :class:`qutip.qobj`
        system Hamiltonian, time-dependent with period `T`

    T : float
        The period of the time-dependence of the hamiltonian.

    args : dictionary
        dictionary with variables required to evaluate H

    options : :class:`qutip.solver`
        options for the ODE solver.

    Returns
    -------

    output : nested list

        A nested list of Floquet modes as kets for each time in `tlist`

    """
    options = copy(options) or Options()
    # truncate tlist to the driving period
    tlist_period = tlist[np.where(tlist <= T)]

    f_modes_table_t = [[] for t in tlist_period]

    opt = options
    opt.rhs_reuse = True
    rhs_clear()

    for n, f_mode in enumerate(f_modes_0):
        output = sesolve(H, f_mode, tlist_period, [], args, opt)
        for t_idx, f_state_t in enumerate(output.states):
            f_modes_table_t[t_idx].append(
                f_state_t * exp(1j * f_energies[n] * tlist_period[t_idx]))

    return f_modes_table_t


def floquet_modes_t_lookup(f_modes_table_t, t, T):
    """
    Lookup the floquet mode at time t in the pre-calculated table of floquet
    modes in the first period of the time-dependence.

    Parameters
    ----------

    f_modes_table_t : nested list of :class:`qutip.qobj` (kets)
        A lookup-table of Floquet modes at times precalculated by
        :func:`qutip.floquet.floquet_modes_table`.

    t : float
        The time for which to evaluate the Floquet modes.

    T : float
        The period of the time-dependence of the hamiltonian.

    Returns
    -------

    output : nested list

        A list of Floquet modes as kets for the time that most closely matching
        the time `t` in the supplied table of Floquet modes.
    """

    # find t_wrap in [0,T] such that t = t_wrap + n * T for integer n
    t_wrap = t - int(t / T) * T

    # find the index in the table that corresponds to t_wrap (= tlist[t_idx])
    t_idx = int(t_wrap / T * len(f_modes_table_t))

    # XXX: might want to give a warning if the cast of t_idx to int discard
    # a significant fraction in t_idx, which would happen if the list of time
    # values isn't perfect matching the driving period
    # if debug: print "t = %f -> t_wrap = %f @ %d of %d" % (t, t_wrap, t_idx,
    # N)

    return f_modes_table_t[t_idx]


def floquet_states(f_modes_t, f_energies, t):
    """
    Evaluate the floquet states at time t given the Floquet modes at that time.

    Parameters
    ----------

    f_modes_t : list of :class:`qutip.qobj` (kets)
        A list of Floquet modes for time :math:`t`.

    f_energies : array
        The Floquet energies.

    t : float
        The time for which to evaluate the Floquet states.

    Returns
    -------

    output : list

        A list of Floquet states for the time :math:`t`.

    """

    return [(f_modes_t[i] * exp(-1j * f_energies[i] * t))
            for i in np.arange(len(f_energies))]


def floquet_states_t(f_modes_0, f_energies, t, H, T, args=None,
                     options=None):
    """
    Evaluate the floquet states at time t given the initial Floquet modes.

    Parameters
    ----------

    f_modes_t : list of :class:`qutip.qobj` (kets)
        A list of initial Floquet modes (for time :math:`t=0`).

    f_energies : array
        The Floquet energies.

    t : float
        The time for which to evaluate the Floquet states.

    H : :class:`qutip.qobj`
        System Hamiltonian, time-dependent with period `T`.

    T : float
        The period of the time-dependence of the hamiltonian.

    args : dictionary
        Dictionary with variables required to evaluate H.

    options : :class:`qutip.solver`
        options for the ODE solver.

    Returns
    -------

    output : list

        A list of Floquet states for the time :math:`t`.

    """

    f_modes_t = floquet_modes_t(f_modes_0, f_energies, t, H, T, args,
                                options=options)
    return [(f_modes_t[i] * exp(-1j * f_energies[i] * t))
            for i in np.arange(len(f_energies))]


def floquet_wavefunction(f_modes_t, f_energies, f_coeff, t):
    """
    Evaluate the wavefunction for a time t using the Floquet state
    decompositon, given the Floquet modes at time `t`.

    Parameters
    ----------

    f_modes_t : list of :class:`qutip.qobj` (kets)
        A list of initial Floquet modes (for time :math:`t=0`).

    f_energies : array
        The Floquet energies.

    f_coeff : array
        The coefficients for Floquet decomposition of the initial wavefunction.

    t : float
        The time for which to evaluate the Floquet states.

    Returns
    -------

    output : :class:`qutip.qobj`

        The wavefunction for the time :math:`t`.

    """
    return sum([f_modes_t[i] * exp(-1j * f_energies[i] * t) * f_coeff[i]
                for i in np.arange(len(f_energies))])


def floquet_wavefunction_t(f_modes_0, f_energies, f_coeff, t, H, T, args=None,
                           options=None):
    """
    Evaluate the wavefunction for a time t using the Floquet state
    decompositon, given the initial Floquet modes.

    Parameters
    ----------

    f_modes_t : list of :class:`qutip.qobj` (kets)
        A list of initial Floquet modes (for time :math:`t=0`).

    f_energies : array
        The Floquet energies.

    f_coeff : array
        The coefficients for Floquet decomposition of the initial wavefunction.

    t : float
        The time for which to evaluate the Floquet states.

    H : :class:`qutip.qobj`
        System Hamiltonian, time-dependent with period `T`.

    T : float
        The period of the time-dependence of the hamiltonian.

    args : dictionary
        Dictionary with variables required to evaluate H.

    Returns
    -------

    output : :class:`qutip.qobj`

        The wavefunction for the time :math:`t`.

    """

    f_states_t = floquet_states_t(f_modes_0, f_energies, t, H, T, args,
                                  options=options)
    return sum([f_states_t[i] * f_coeff[i]
                for i in np.arange(len(f_energies))])


def floquet_state_decomposition(f_states, f_energies, psi):
    r"""
    Decompose the wavefunction `psi` (typically an initial state) in terms of
    the Floquet states, :math:`\psi = \sum_\alpha c_\alpha \psi_\alpha(0)`.

    Parameters
    ----------

    f_states : list of :class:`qutip.qobj` (kets)
        A list of Floquet modes.

    f_energies : array
        The Floquet energies.

    psi : :class:`qutip.qobj`
        The wavefunction to decompose in the Floquet state basis.

    Returns
    -------

    output : array

        The coefficients :math:`c_\alpha` in the Floquet state decomposition.

    """
    # [:1,:1][0, 0] patch around scipy 1.3.0 bug
    return [(f_states[i].dag() * psi).data[:1, :1][0, 0]
            for i in np.arange(len(f_energies))]


def fsesolve(H, psi0, tlist, e_ops=[], T=None, args={}, Tsteps=100,
             options_modes=None):
    """
    Solve the Schrodinger equation using the Floquet formalism.

    Parameters
    ----------

    H : :class:`qutip.qobj.Qobj`
        System Hamiltonian, time-dependent with period `T`.

    psi0 : :class:`qutip.qobj`
        Initial state vector (ket).

    tlist : *list* / *array*
        list of times for :math:`t`.

    e_ops : list of :class:`qutip.qobj` / callback function
        list of operators for which to evaluate expectation values. If this
        list is empty, the state vectors for each time in `tlist` will be
        returned instead of expectation values.

    T : float
        The period of the time-dependence of the hamiltonian.

    args : dictionary
        Dictionary with variables required to evaluate H.

    Tsteps : integer
        The number of time steps in one driving period for which to
        precalculate the Floquet modes. `Tsteps` should be an even number.

    options_modes : :class:`qutip.solver`
        options for the ODE solver.

    Returns
    -------

    output : :class:`qutip.solver.Result`

        An instance of the class :class:`qutip.solver.Result`, which
        contains either an *array* of expectation values or an array of
        state vectors, for the times specified by `tlist`.
    """

    if not T:
        # assume that tlist span exactly one period of the driving
        T = tlist[-1]

    if options_modes is None:
        options_modes_table = Options()
    else:
        options_modes_table = options_modes

    # find the floquet modes for the time-dependent hamiltonian
    f_modes_0, f_energies = floquet_modes(H, T, args,
                                          options=options_modes)

    # calculate the wavefunctions using the from the floquet modes
    f_modes_table_t = floquet_modes_table(f_modes_0, f_energies,
                                          np.linspace(0, T, Tsteps + 1),
                                          H, T, args,
                                          options=options_modes_table)

    # setup Result for storing the results
    output = Result()
    output.times = tlist
    output.solver = "fsesolve"

    if isinstance(e_ops, FunctionType):
        output.num_expect = 0
        expt_callback = True

    elif isinstance(e_ops, list):

        output.num_expect = len(e_ops)
        expt_callback = False

        if output.num_expect == 0:
            output.states = []
        else:
            output.expect = []
            for op in e_ops:
                if op.isherm:
                    output.expect.append(np.zeros(len(tlist)))
                else:
                    output.expect.append(np.zeros(len(tlist), dtype=complex))

    else:
        raise TypeError("e_ops must be a list Qobj or a callback function")

    psi0_fb = psi0.transform(f_modes_0)
    for t_idx, t in enumerate(tlist):
        f_modes_t = floquet_modes_t_lookup(f_modes_table_t, t, T)
        f_states_t = floquet_states(f_modes_t, f_energies, t)
        psi_t = psi0_fb.transform(f_states_t, True)

        if expt_callback:
            # use callback method
            e_ops(t, psi_t)
        else:
            # calculate all the expectation values, or output psi if
            # no expectation value operators where defined
            if output.num_expect == 0:
                output.states.append(Qobj(psi_t))
            else:
                for e_idx, e in enumerate(e_ops):
                    output.expect[e_idx][t_idx] = expect(e, psi_t)

    return output


def floquet_master_equation_rates(f_modes_0, f_energies, c_op, H, T,
                                  args, J_cb, w_th, kmax=5,
                                  f_modes_table_t=None):
    """
    Calculate the rates and matrix elements for the Floquet-Markov master
    equation.

    Parameters
    ----------

    f_modes_0 : list of :class:`qutip.qobj` (kets)
        A list of initial Floquet modes.

    f_energies : array
        The Floquet energies.

    c_op : :class:`qutip.qobj`
        The collapse operators describing the dissipation.

    H : :class:`qutip.qobj`
        System Hamiltonian, time-dependent with period `T`.

    T : float
        The period of the time-dependence of the hamiltonian.

    args : dictionary
        Dictionary with variables required to evaluate H.

    J_cb : callback functions
        A callback function that computes the noise power spectrum, as
        a function of frequency, associated with the collapse operator `c_op`.

    w_th : float
        The temperature in units of frequency.

    k_max : int
        The truncation of the number of sidebands (default 5).

    f_modes_table_t : nested list of :class:`qutip.qobj` (kets)
        A lookup-table of Floquet modes at times precalculated by
        :func:`qutip.floquet.floquet_modes_table` (optional).

    options : :class:`qutip.solver`
        options for the ODE solver.

    Returns
    -------

    output : list

        A list (Delta, X, Gamma, A) containing the matrices Delta, X, Gamma
        and A used in the construction of the Floquet-Markov master equation.

    """

    N = len(f_energies)
    M = 2 * kmax + 1

    omega = (2 * pi) / T

    Delta = np.zeros((N, N, M))
    X = np.zeros((N, N, M), dtype=complex)
    Gamma = np.zeros((N, N, M))
    A = np.zeros((N, N))

    nT = 100
    dT = T / nT
    tlist = np.arange(dT, T + dT / 2, dT)

    if f_modes_table_t is None:
        f_modes_table_t = floquet_modes_table(f_modes_0, f_energies,
                                              np.linspace(0, T, nT + 1), H, T,
                                              args)

    for t in tlist:
        # TODO: repeated invocations of floquet_modes_t is
        # inefficient...  make a and b outer loops and use the mesolve
        # instead of the propagator.

        f_modes_t = np.hstack([f.full() for f in floquet_modes_t_lookup(
            f_modes_table_t, t, T)])
        FF = f_modes_t.T.conj() @ c_op.full() @ f_modes_t
        phi = exp(-1j * np.arange(-kmax, kmax+1) * omega * t)
        X += (dT / T) * np.einsum("ij,k->ijk", FF, phi)

    Heaviside = lambda x: ((np.sign(x) + 1) / 2.0)
    for a in range(N):
        for b in range(N):
            k_idx = 0
            for k in range(-kmax, kmax + 1, 1):
                Delta[a, b, k_idx] = f_energies[a] - f_energies[b] + k * omega
                Gamma[a, b, k_idx] = 2 * pi * Heaviside(Delta[a, b, k_idx]) * \
                    J_cb(Delta[a, b, k_idx]) * abs(X[a, b, k_idx]) ** 2
                k_idx += 1

    for a in range(N):
        for b in range(N):
            for k in range(-kmax, kmax + 1, 1):
                k1_idx = k + kmax
                k2_idx = -k + kmax
                A[a, b] += Gamma[a, b, k1_idx] + \
                    n_thermal(abs(Delta[a, b, k1_idx]), w_th) * \
                    (Gamma[a, b, k1_idx] + Gamma[b, a, k2_idx])

    return Delta, X, Gamma, A


def floquet_collapse_operators(A):
    """
    Construct collapse operators corresponding to the Floquet-Markov
    master-equation rate matrix `A`.

    .. note::

        Experimental.

    """
    c_ops = []

    N, M = np.shape(A)

    #
    # Here we really need a master equation on Bloch-Redfield form, or perhaps
    # we can use the Lindblad form master equation with some rotating frame
    # approximations? ...
    #
    for a in range(N):
        for b in range(N):
            if a != b and abs(A[a, b]) > 0.0:
                # only relaxation terms included...
                c_ops.append(sqrt(A[a, b]) * projection(N, a, b))

    return c_ops


def floquet_master_equation_tensor(Alist, f_energies):
    """
    Construct a tensor that represents the master equation in the floquet
    basis (with constant Hamiltonian and collapse operators).

    Simplest RWA approximation [Grifoni et al, Phys.Rep. 304 229 (1998)]

    Parameters
    ----------

    Alist : list
        A list of Floquet-Markov master equation rate matrices.

    f_energies : array
        The Floquet energies.

    Returns
    -------

    output : array

        The Floquet-Markov master equation tensor `R`.

    """

    if isinstance(Alist, list):
        # Alist can be a list of rate matrices corresponding
        # to different operators that couple to the environment
        N, M = np.shape(Alist[0])
    else:
        # or a simple rate matrix, in which case we put it in a list
        Alist = [Alist]
        N, M = np.shape(Alist[0])

    Rdata_lil = scipy.sparse.lil_matrix((N * N, N * N), dtype=complex)

    AsumList = [np.sum(A, axis=1) for A in Alist]

    for k in range(len(Alist)):
        for i in range(N):
            Rdata_lil[i+N*i, i+N*i] -= -Alist[k][i, i] + AsumList[k][i]
            for j in range(i+1, N):
                Rdata_lil[i+N*i, j+N*j] += Alist[k][j, i]
                Rdata_lil[j+N*j, i+N*i] += Alist[k][i, j]
                a_term = -(1/2)*(AsumList[k][i] + AsumList[k][j])
                Rdata_lil[i+N*j, i+N*j] += a_term
                Rdata_lil[j+N*i, j+N*i] += a_term

    return Qobj(Rdata_lil, dims=[[N, N], [N, N]], shape=(N*N, N*N))


def floquet_master_equation_steadystate(H, A):
    """
    Returns the steadystate density matrix (in the floquet basis!) for the
    Floquet-Markov master equation.
    """
    c_ops = floquet_collapse_operators(A)
    rho_ss = steadystate(H, c_ops)
    return rho_ss


def floquet_basis_transform(f_modes, f_energies, rho0):
    """
    Make a basis transform that takes rho0 from the floquet basis to the
    computational basis.
    """
    return rho0.transform(f_modes, True)

# -----------------------------------------------------------------------------
# Floquet-Markov master equation
#
#


def floquet_markov_mesolve(
    R, rho0, tlist, e_ops, options=None, floquet_basis=True,
    f_modes_0=None, f_modes_table_t=None, f_energies=None, T=None,
):
    """
    Solve the dynamics for the system using the Floquet-Markov master equation.

    .. note::

        It is important to understand in which frame and basis the results
        are returned here.

    Parameters
    ----------

    R : array
        The Floquet-Markov master equation tensor `R`.

    rho0 : :class:`qutip.qobj`
        Initial density matrix.  If ``f_modes_0`` is not passed, this density
        matrix is assumed to be in the Floquet picture.

    tlist : *list* / *array*
        list of times for :math:`t`.

    e_ops : list of :class:`qutip.qobj` / callback function
        list of operators for which to evaluate expectation values.

    options : :class:`qutip.solver`
        options for the ODE solver.

    floquet_basis: bool, True
        If ``True``, states and expectation values will be returned in the
        Floquet basis.  If ``False``, a transformation will be made to the
        computational basis; this will be in the lab frame if
        ``f_modes_table``, ``T` and ``f_energies`` are all supplied, or the
        interaction picture (defined purely be f_modes_0) if they are not.

    f_modes_0 : list of :class:`qutip.qobj` (kets), optional
        A list of initial Floquet modes, used to transform the given starting
        density matrix into the Floquet basis.  If this is not passed, it is
        assumed that ``rho`` is already in the Floquet basis.

    f_modes_table_t : nested list of :class:`qutip.qobj` (kets), optional
        A lookup-table of Floquet modes at times precalculated by
        :func:`qutip.floquet.floquet_modes_table`.  Necessary if
        ``floquet_basis`` is ``False`` and the transformation should be made
        back to the lab frame.

    f_energies : array_like of float, optional
        The precalculated Floquet quasienergies.  Necessary if
        ``floquet_basis`` is ``False`` and the transformation should be made
        back to the lab frame.

    T : float, optional
        The time period of driving.  Necessary if ``floquet_basis`` is
        ``False`` and the transformation should be made back to the lab frame.

    Returns
    -------

    output : :class:`qutip.solver.Result`
        An instance of the class :class:`qutip.solver.Result`, which
        contains either an *array* of expectation values or an array of
        state vectors, for the times specified by `tlist`.
    """
    opt = options or Options()
    if opt.tidy:
        R.tidyup()
    rho0 = rho0.proj() if rho0.isket else rho0

    # Prepare output object.
    dt = tlist[1] - tlist[0]
    output = Result()
    output.solver = "fmmesolve"
    output.times = tlist
    if isinstance(e_ops, FunctionType):
        expt_callback = True
        store_states = opt.store_states or False
    else:
        expt_callback = False
        try:
            e_ops = list(e_ops)
        except TypeError:
            raise TypeError("`e_ops` must be iterable or a function") from None
        n_expt_op = len(e_ops)
        if n_expt_op == 0:
            store_states = True
        else:
            output.expect = []
            output.num_expect = n_expt_op
            for op in e_ops:
                dtype = np.float64 if op.isherm else np.complex128
                output.expect.append(np.zeros(len(tlist), dtype=dtype))
        store_states = opt.store_states or (n_expt_op == 0)
    if store_states:
        output.states = []

    # Choose which frame transformations should be done on the initial and
    # evolved states.
    lab_lookup = [f_modes_table_t, f_energies, T]
    if (
        any(x is None for x in lab_lookup)
        and not all(x is None for x in lab_lookup)
    ):
        warnings.warn(
            "if transformation back to the computational basis in the lab"
            "frame is desired, all of `f_modes_t`, `f_energies` and `T` must"
            "be supplied."
        )
        f_modes_table_t = f_energies = T = None

    # Initial state.
    if f_modes_0 is not None:
        rho0 = rho0.transform(f_modes_0)

    # Evolved states.
    if floquet_basis:
        def transform(rho, t):
            return rho
    elif f_modes_table_t is not None:
        # Lab frame, computational basis.
        def transform(rho, t):
            f_modes_t = floquet_modes_t_lookup(f_modes_table_t, t, T)
            f_states_t = floquet_states(f_modes_t, f_energies, t)
            return rho.transform(f_states_t, True)
    elif f_modes_0 is not None:
        # Interaction picture, computational basis.
        def transform(rho, t):
            return rho.transform(f_modes_0, False)
    else:
        raise ValueError(
            "cannot transform out of the Floquet basis without some knowledge "
            "of the Floquet modes.  Pass `f_modes_0`, or all of `f_modes_t`, "
            "`f_energies` and `T`."
        )

    # Setup integrator.
    initial_vector = mat2vec(rho0.full())
    r = scipy.integrate.ode(cy_ode_rhs)
    r.set_f_params(R.data.data, R.data.indices, R.data.indptr)
    r.set_integrator('zvode', method=opt.method, order=opt.order,
                     atol=opt.atol, rtol=opt.rtol, max_step=opt.max_step)
    r.set_initial_value(initial_vector, tlist[0])

    # Main evolution loop.
    for t_idx, t in enumerate(tlist):
        if not r.successful():
            break
        rho = transform(Qobj(vec2mat(r.y), rho0.dims, rho0.shape), t)
        if expt_callback:
            e_ops(t, rho)
        else:
            for m, e_op in enumerate(e_ops):
                output.expect[m][t_idx] = expect(e_op, rho)
        if store_states:
            output.states.append(rho)
        r.integrate(r.t + dt)
    return output

# -----------------------------------------------------------------------------
# Solve the Floquet-Markov master equation
#
#


def fmmesolve(H, rho0, tlist, c_ops=[], e_ops=[], spectra_cb=[], T=None,
              args={}, options=Options(), floquet_basis=True, kmax=5,
              _safe_mode=True, options_modes=None):
    """
    Solve the dynamics for the system using the Floquet-Markov master equation.

    .. note::

        This solver currently does not support multiple collapse operators.

    Parameters
    ----------

    H : :class:`qutip.qobj`
        system Hamiltonian.

    rho0 / psi0 : :class:`qutip.qobj`
        initial density matrix or state vector (ket).

    tlist : *list* / *array*
        list of times for :math:`t`.

    c_ops : list of :class:`qutip.qobj`
        list of collapse operators.

    e_ops : list of :class:`qutip.qobj` / callback function
        list of operators for which to evaluate expectation values.

    spectra_cb : list callback functions
        List of callback functions that compute the noise power spectrum as
        a function of frequency for the collapse operators in `c_ops`.

    T : float
        The period of the time-dependence of the hamiltonian. The default value
        'None' indicates that the 'tlist' spans a single period of the driving.

    args : *dictionary*
        dictionary of parameters for time-dependent Hamiltonians and
        collapse operators.

        This dictionary should also contain an entry 'w_th', which is
        the temperature of the environment (if finite) in the
        energy/frequency units of the Hamiltonian.  For example, if
        the Hamiltonian written in units of 2pi GHz, and the
        temperature is given in K, use the following conversion

        >>> temperature = 25e-3 # unit K # doctest: +SKIP
        >>> h = 6.626e-34 # doctest: +SKIP
        >>> kB = 1.38e-23 # doctest: +SKIP
        >>> args['w_th'] = temperature * (kB / h) * 2 * pi * 1e-9 \
            #doctest: +SKIP

    options : :class:`qutip.solver`
        options for the ODE solver. For solving the master equation.

    floquet_basis : bool
        Will return results in Floquet basis or computational basis
        (optional).

    k_max : int
        The truncation of the number of sidebands (default 5).

    options_modes : :class:`qutip.solver`
        options for the ODE solver. For computing Floquet modes.

    Returns
    -------

    output : :class:`qutip.solver`

        An instance of the class :class:`qutip.solver`, which contains either
        an *array* of expectation values for the times specified by `tlist`.
    """

    if _safe_mode:
        _solver_safety_check(H, rho0, c_ops, e_ops, args)

    if options_modes is None:
        options_modes_table = Options()
    else:
        options_modes_table = options_modes

    if T is None:
        T = max(tlist)

    if len(spectra_cb) == 0:
        # add white noise callbacks if absent
        spectra_cb = [lambda w: 1.0] * len(c_ops)

    f_modes_0, f_energies = floquet_modes(H, T, args,
                                          options=options_modes)

    f_modes_table_t = floquet_modes_table(f_modes_0, f_energies,
                                          np.linspace(0, T, 500 + 1),
                                          H, T, args,
                                          options=options_modes_table)

    # get w_th from args if it exists
    if 'w_th' in args:
        w_th = args['w_th']
    else:
        w_th = 0

    # TODO: loop over input c_ops and spectra_cb, calculate one R for each set

    # calculate the rate-matrices for the floquet-markov master equation
    Delta, X, Gamma, Amat = floquet_master_equation_rates(
        f_modes_0, f_energies, c_ops[0], H, T, args, spectra_cb[0],
        w_th, kmax, f_modes_table_t)

    # the floquet-markov master equation tensor
    R = floquet_master_equation_tensor(Amat, f_energies)

    return floquet_markov_mesolve(R, rho0, tlist, e_ops,
                                  options=options,
                                  floquet_basis=floquet_basis,
                                  f_modes_0=f_modes_0,
                                  f_modes_table_t=f_modes_table_t,
                                  T=T,
                                  f_energies=f_energies)



from qutip.sparse import eigh
from qutip import krylovsolve, sesolve
from qutip.solver import Options, Result
from qutip import tensor, Qobj, basis, expect, ket
from qutip import sigmax, sigmay, sigmaz, qeye, jmat
from qutip.qip.operations import x_gate, y_gate, z_gate
from qutip import rand_herm, rand_ket, rand_unitary_haar, num, destroy, create
import os
import math
import types
import pytest
from scipy.linalg import expm
import numpy as np

np.random.seed(0)


@pytest.fixture(params=[
    pytest.param(100, id="small dim"),
    pytest.param(500, id="intermediate dim"),
    pytest.param(1000, id="large dim"),
])
def dimensions(request):
    return request.param


@pytest.fixture(params=[
    pytest.param(np.linspace(0, 5, 200), id="normal tlist"),
    pytest.param([2], id="single element tlist"),
    pytest.param([], id="empty tlist"),
])
def tlists(request):
    return request.param


@pytest.fixture(params=[
    pytest.param((ket([1, 0, 0, 0]), 0.5, 0, 1), id="eigenstate"),
    pytest.param((ket([0, 0, 1, 0]), 1.0, 1, 0),
                 id="magnetization subspace state XXZ model"),
])
def happy_breakdown_parameters(request):
    return request.param


def h_sho(dim):
    U = rand_unitary_haar(dim)
    H = U * (num(dim) + 0.5) * U.dag()
    H = H / np.linalg.norm(H.full(), ord=2)
    return H


def h_ising_transverse(
    N: int, hx: float, hz: float, Jx: float, Jy: float, Jz: float
):

    hx = hx * np.ones(N)
    hz = hz * np.ones(N)
    Jx, Jy, Jz = Jx * np.ones(N), Jy * np.ones(N), Jz * np.ones(N)

    sx_list = [x_gate(N, i) for i in range(N)]
    sy_list = [y_gate(N, i) for i in range(N)]
    sz_list = [z_gate(N, i) for i in range(N)]

    # construct the hamiltonian
    H = 0

    # energy splitting terms
    for n in range(N):
        H += hz[n] * sz_list[n]
        H += hx[n] * sx_list[n]

        if n < N-1:
            # interaction terms
            H += -Jx[n] * sx_list[n] * sx_list[n + 1]
            H += -Jy[n] * sy_list[n] * sy_list[n + 1]
            H += -Jz[n] * sz_list[n] * sz_list[n + 1]

    return H


def create_test_e_ops(e_ops_type, dim):
    """Creates e_ops used for testing give H and dim."""
    if e_ops_type == "[c]":
        e_ops = [lambda t, psi: expect(num(dim), psi)]
    if e_ops_type == "[q]":
        e_ops = [jmat((dim - 1)/2, "x")]
    if e_ops_type == "[c, c]":
        e_ops = [lambda t, psi: expect(num(dim), psi),
                 lambda t, psi: expect(num(dim)/2, psi)]
    if e_ops_type == "[c, q]":
        e_ops = [lambda t, psi: expect(num(dim), psi),
                 jmat((dim - 1) / 2.0, "x")]
    if e_ops_type == "[q, q]":
        e_ops = [jmat((dim - 1) / 2.0, "x"),
                 jmat((dim - 1) / 2.0, "y"),
                 jmat((dim - 1) / 2.0, "z")]
    return e_ops


def err_psi(psi_a, psi_b):
    """Error between to kets."""
    err = 1 - np.abs(psi_a.overlap(psi_b)) ** 2
    return err


def expect_value(e_ops, res_1, tlist):
    """Calculates expectation values of results object."""
    if isinstance(e_ops, types.FunctionType):
        expect_values = [
            e_ops(t, state) for (t, state) in zip(tlist, res_1.states)
        ]
    else:
        expect_values = [expect(e_ops, state) for state in res_1.states]
    return expect_values


def assert_err_states_less_than_tol(res_1, res_2, tol):
    """Asserts error of states of two results less than tol."""

    err_states = [err_psi(psi_1, psi_2) for (psi_1, psi_2)
                  in zip(res_1.states, res_2.states)]
    for err in err_states:
        assert err <= tol,\
            f"err in states {err} is > than tolerance {tol}."


def assert_err_expect_less_than_tol(exp_1, exp_2, tol):
    """Asserts error of expect values of two results less than tol."""
    err_expect = [np.abs(ex_1 - ex_2) for (ex_1, ex_2)
                  in zip(exp_1, exp_2)]

    for err in err_expect:
        assert err <= tol, \
            f"err in expect values {err} is > than tol {tol}."


def exactsolve(H, psi0, tlist):
    """Calculates exact solution by direct diagonalization."""

    dims = psi0.dims
    H = H.full()
    psi0 = psi0.full()

    eigenvalues, eigenvectors = eigh(H)

    psi_base_diag = np.matmul(eigenvectors.conj().T, psi0)
    U = np.exp(np.outer(-1j * eigenvalues, tlist))
    psi_list = np.matmul(
        eigenvectors, np.multiply(U, psi_base_diag.reshape([-1, 1]))
    )
    exact_results = Result()
    exact_results.states = [Qobj(state, dims=dims) for state in psi_list.T]

    return exact_results


class TestKrylovSolve:
    """
    A test class for the QuTiP Krylov Approximation method Solver.
    """

    def check_sparse_vs_dense(self, output_sparse, output_dense, tol=1e-5):
        "simple check of errors between two outputs states"

        assert_err_states_less_than_tol(res_1=output_sparse,
                                        res_2=output_dense,
                                        tol=tol)

    @pytest.mark.parametrize("density,dim", [(0.1, 512), (0.9, 800)],
                             ids=["sparse H check", "dense H check"])
    def test_01_check_sparse_vs_non_sparse_with_density_H(
        self, density, dim, krylov_dim=20
    ):
        "krylovsolve: comparing sparse vs non sparse."
        psi0 = rand_ket(dim)
        H_sparse = rand_herm(dim, density=0.1)
        tlist = np.linspace(0, 10, 200)

        output_sparse = krylovsolve(
            H_sparse, psi0, tlist, krylov_dim, sparse=True
        )
        output_dense = krylovsolve(
            H_sparse, psi0, tlist, krylov_dim, sparse=False
        )

        self.check_sparse_vs_dense(output_sparse, output_dense)

    def simple_check_states_e_ops(
        self,
        H,
        psi0,
        tlist,
        tol=1e-5,
        tol2=1e-4,
        krylov_dim=25,
        square_hamiltonian=True,
    ):
        """
        Compare integrated evolution with sesolve and exactsolve result.
        """

        options = Options(store_states=True)

        e_ops = [
            jmat((H.shape[0] - 1) / 2.0, "x"),
            jmat((H.shape[0] - 1) / 2.0, "y"),
            jmat((H.shape[0] - 1) / 2.0, "z"),
        ]
        if not square_hamiltonian:
            _e_ops = []
            for op in e_ops:
                op2 = op.copy()
                op2.dims = H.dims
                _e_ops.append(op2)
            e_ops = _e_ops

        output = krylovsolve(
            H, psi0, tlist, krylov_dim, e_ops=e_ops, options=options
        )
        output_ss = sesolve(H, psi0, tlist, e_ops=e_ops, options=options)
        output_exact = exactsolve(H, psi0, tlist)

        assert_err_states_less_than_tol(res_1=output,
                                        res_2=output_ss, tol=tol)

        assert_err_states_less_than_tol(res_1=output,
                                        res_2=output_exact, tol=tol)

        # for the operators, test against exactsolve for accuracy
        for i in range(len(e_ops)):

            output_exact.expect = expect_value(e_ops[i], output_exact, tlist)

            assert_err_expect_less_than_tol(exp_1=output.expect[i],
                                            exp_2=output_exact.expect,
                                            tol=tol)

    def test_02_simple_check_states_e_ops_H_random(self):
        "krylovsolve: states with const H random"
        dim = 512
        psi0 = rand_ket(dim)
        H = rand_herm(dim)
        tlist = np.linspace(0, 10, 200)

        self.simple_check_states_e_ops(H, psi0, tlist)

    def test_03_simple_check_states_e_ops_H_ising_transverse(self):
        "krylovsolve: states with const H Ising Transverse Field"
        N = 8
        dim = 2**N
        H = h_ising_transverse(N, hx=0.1, hz=0.5, Jx=1, Jy=0, Jz=1)
        _dims = H.dims
        _dims2 = [1] * N
        psi0 = rand_ket(dim, dims=[_dims[0], _dims2])
        tlist = np.linspace(0, 20, 200)

        self.simple_check_states_e_ops(
            H, psi0, tlist, square_hamiltonian=False)

    def test_04_simple_check_states_e_ops_H_sho(self):
        "krylovsolve: states with const H SHO"
        dim = 100
        psi0 = rand_ket(dim)
        H = h_sho(dim)
        tlist = np.linspace(0, 20, 200)

        self.simple_check_states_e_ops(H, psi0, tlist)

    def check_e_ops_none(
        self, H, psi0, tlist, dim, krylov_dim=30, tol=1e-5, tol2=1e-4
    ):
        "Check input possibilities when e_ops=None"

        krylov_outputs = krylovsolve(H, psi0, tlist, krylov_dim, e_ops=None)

        try:
            sesolve_outputs = sesolve(H, psi0, tlist, e_ops=None)
        except IndexError:  # if tlist=[], sesolve breaks but krylov doesn't
            pass

        if len(tlist) > 1:
            assert_err_states_less_than_tol(
                krylov_outputs, sesolve_outputs, tol)
        elif len(tlist) == 1:
            assert krylov_outputs.states == sesolve_outputs.states
        else:
            assert krylov_outputs.states == []

    def test_05_check_e_ops_none(self, dimensions, tlists):
        "krylovsolve: check e_ops=None inputs with random H different tlists."
        psi0 = rand_ket(dimensions)
        H = rand_herm(dimensions, density=0.5)
        self.check_e_ops_none(H, psi0, tlists, dimensions)

    def check_e_ops_callable(
        self,
        H,
        psi0,
        tlist,
        dim,
        krylov_dim=35,
        tol=1e-5,
        square_hamiltonian=True,
    ):
        "Check input possibilities when e_ops=callable"

        def e_ops(t, psi): return expect(num(dim), psi)

        if not square_hamiltonian:
            H.dims = [[H.shape[0]], [H.shape[0]]]
            psi0.dims = [[H.shape[0]], [1]]

        krylov_outputs = krylovsolve(H, psi0, tlist, krylov_dim, e_ops=e_ops)
        exact_output = exactsolve(H, psi0, tlist)
        exact_output.expect = expect_value(e_ops=e_ops,
                                           res_1=exact_output,
                                           tlist=tlist)

        try:
            sesolve_outputs = sesolve(H, psi0, tlist, e_ops=e_ops)
        except IndexError:  # if tlist=[], sesolve breaks but krylov doesn't
            pass

        if len(tlist) > 1:

            assert len(krylov_outputs.expect) == len(
                sesolve_outputs.expect
            ), "shape of outputs between krylov and sesolve differs"

            assert_err_expect_less_than_tol(exp_1=krylov_outputs.expect,
                                            exp_2=exact_output.expect,
                                            tol=tol)
        elif len(tlist) == 1:
            assert (
                np.abs(krylov_outputs.expect[0] - sesolve_outputs.expect[0])
                <= 1e-7
            ), "krylov and sesolve outputs differ for len(tlist)=1"
        else:
            assert krylov_outputs.states == []

    def test_06_check_e_ops_callable(self, dimensions, tlists):
        "krylovsolve: check e_ops=call inputs with random H different tlists."
        psi0 = rand_ket(dimensions)
        H = rand_herm(dimensions, density=0.5)
        self.check_e_ops_callable(H, psi0, tlists, dimensions)

    def check_e_ops_list_single_operator(
        self,
        e_ops,
        H,
        psi0,
        tlist,
        dim,
        krylov_dim=35,
        tol=1e-5,
        square_hamiltonian=True,
    ):
        "Check input possibilities when e_ops=[callable | qobj]"

        if not square_hamiltonian:
            H.dims = [[H.shape[0]], [H.shape[0]]]
            psi0.dims = [[H.shape[0]], [1]]

        krylov_outputs = krylovsolve(H, psi0, tlist, krylov_dim, e_ops=e_ops)
        exact_output = exactsolve(H, psi0, tlist)
        exact_output.expect = expect_value(e_ops[0], exact_output, tlist)

        try:
            sesolve_outputs = sesolve(H, psi0, tlist, e_ops=e_ops)
        except IndexError:  # if tlist=[], sesolve breaks but krylov doesn't
            pass

        if len(tlist) > 1:

            assert len(krylov_outputs.expect) == len(
                sesolve_outputs.expect
            ), "shape of outputs between krylov and sesolve differs"

            assert_err_expect_less_than_tol(exp_1=krylov_outputs.expect[0],
                                            exp_2=exact_output.expect,
                                            tol=tol)

        elif len(tlist) == 1:
            assert (
                np.abs(krylov_outputs.expect[0] - sesolve_outputs.expect[0])
                <= tol
            ), "expect outputs from krylovsolve and sesolve are not equal"
        else:
            assert krylov_outputs.states == []

    @pytest.mark.parametrize("e_ops_type", [("[c]"), ("[q]")], ids=["[c]", "[q]"])
    def test_07_check_e_ops_list_single_callable(self, e_ops_type, dimensions, tlists):
        "krylovsolve: check e_ops=[call | qobj] random H different tlists."
        psi0 = rand_ket(dimensions)
        H = rand_herm(dimensions, density=0.5)
        e_ops = create_test_e_ops(e_ops_type, dimensions)
        self.check_e_ops_list_single_operator(
            e_ops, H, psi0, tlists, dimensions)

    def check_e_ops_mixed_list(
        self, e_ops, H, psi0, tlist, dim, krylov_dim=35, tol=1e-5,
    ):
        "Check input possibilities when e_ops=[call | qobj] and len(e_ops) > 1"

        krylov_outputs = krylovsolve(H, psi0, tlist, krylov_dim, e_ops=e_ops)
        exact_output = exactsolve(H, psi0, tlist)

        try:
            sesolve_outputs = sesolve(H, psi0, tlist, e_ops=e_ops)
        except IndexError:  # if tlist=[], sesolve breaks but krylov doesn't
            pass

        if len(tlist) > 1:

            assert len(krylov_outputs.expect) == len(
                sesolve_outputs.expect
            ), "shape of outputs between krylov and sesolve differs"

            for idx, k_expect in enumerate(krylov_outputs.expect):
                exact_output.expect = expect_value(
                    e_ops[idx], exact_output, tlist)
                assert_err_expect_less_than_tol(exp_1=k_expect,
                                                exp_2=exact_output.expect,
                                                tol=tol)
        elif len(tlist) == 1:
            assert len(krylov_outputs.expect) == len(
                sesolve_outputs.expect
            ), "shape of outputs between krylov and sesolve differs"
            for k_expect, ss_expect in zip(
                krylov_outputs.expect, sesolve_outputs.expect
            ):
                assert (
                    np.abs(k_expect - ss_expect) <= tol
                ), "expect outputs from krylovsolve and sesolve are not equal"
        else:
            assert krylov_outputs.states == []

    @pytest.mark.parametrize("e_ops_type",
                             [("[c, c]"), ("[c, q]"), ("[q, q]")],
                             ids=["[c, c]", "[c, q]", "[q, q]"])
    def test_08_check_e_ops_mixed_list(self, e_ops_type, dimensions, tlists):
        "krylovsolve: check e_ops=[call | qobj] with len(e_ops)>1 for"
        "random H different tlists."
        psi0 = rand_ket(dimensions)
        H = rand_herm(dimensions, density=0.5)
        e_ops = create_test_e_ops(e_ops_type, dimensions)
        self.check_e_ops_mixed_list(e_ops, H, psi0, tlists, dimensions)

    def test_9_happy_breakdown_simple(self, happy_breakdown_parameters):
        "krylovsolve: check simple at happy breakdowns"
        psi0, hz, Jx, Jz = happy_breakdown_parameters
        krylov_dim = 12
        N = 4
        dim = 2**N
        H = h_ising_transverse(N, hx=0, hz=hz, Jx=Jx, Jy=0, Jz=Jz)
        tlist = np.linspace(0, 20, 200)
        self.simple_check_states_e_ops(
            H, psi0, tlist, krylov_dim=krylov_dim, square_hamiltonian=False
        )

    def test_10_happy_breakdown_e_ops_none(self, happy_breakdown_parameters):
        "krylovsolve: check e_ops=None at happy breakdowns"
        psi0, hz, Jx, Jz = happy_breakdown_parameters
        krylov_dim = 12
        N = 4
        dim = 2**N
        H = h_ising_transverse(N, hx=0, hz=hz, Jx=Jx, Jy=0, Jz=Jz)
        tlist = np.linspace(0, 20, 200)
        self.check_e_ops_none(
            H, psi0, tlist, dim, krylov_dim=krylov_dim
        )

    def test_11_happy_breakdown_e_ops_callable(self, happy_breakdown_parameters):
        "krylovsolve: check e_ops=callable at happy breakdowns"
        psi0, hz, Jx, Jz = happy_breakdown_parameters
        krylov_dim = 12
        N = 4
        dim = 2**N
        H = h_ising_transverse(N, hx=0, hz=hz, Jx=Jx, Jy=0, Jz=Jz)
        def e_ops(t, psi): return expect(num(dim), psi)
        tlist = np.linspace(0, 20, 200)

        self.check_e_ops_callable(
            H,
            psi0,
            tlist,
            dim,
            krylov_dim=krylov_dim,
            square_hamiltonian=False,
        )

    def test_12_happy_breakdown_e_ops_list_single_callable(self, happy_breakdown_parameters):
        "krylovsolve: check e_ops=[callable] at happy breakdowns"
        psi0, hz, Jx, Jz = happy_breakdown_parameters
        krylov_dim = 12
        N = 4
        dim = 2**N
        H = h_ising_transverse(N, hx=0, hz=hz, Jx=Jx, Jy=0, Jz=Jz)
        e_ops = [lambda t, psi: expect(num(dim), psi)]

        tlist = np.linspace(0, 20, 200)
        self.check_e_ops_list_single_operator(
            e_ops,
            H,
            psi0,
            tlist,
            dim,
            krylov_dim=krylov_dim,
            square_hamiltonian=False,
        )


def test_13_krylovsolve_bad_krylov_dim(dim=15, krylov_dim=20):
    """Check errors from bad krylov dimension inputs."""
    H = rand_herm(dim)
    psi0 = basis(dim, 0)
    tlist = np.linspace(0, 1, 100)
    with pytest.raises(ValueError) as exc:
        krylovsolve(H, psi0, tlist, krylov_dim)



import pytest
import numpy as np
import scipy.linalg
import scipy.sparse

import qutip
if qutip.settings.has_mkl:
    from qutip._mkl.spsolve import mkl_splu, mkl_spsolve

pytestmark = [
    pytest.mark.skipif(not qutip.settings.has_mkl,
                       reason='MKL extensions not found.'),
]


class Test_spsolve:
    def test_single_rhs_vector_real(self):
        Adense = np.array([[0, 1, 1],
                           [1, 0, 1],
                           [0, 0, 1]])
        As = scipy.sparse.csr_matrix(Adense)
        np.random.seed(1234)
        x = np.random.randn(3)
        b = As * x
        x2 = mkl_spsolve(As, b, verbose=True)
        np.testing.assert_allclose(x, x2)

    def test_single_rhs_vector_complex(self):
        A = qutip.rand_herm(10)
        x = qutip.rand_ket(10).full()
        b = A.full() @ x
        y = mkl_spsolve(A.data, b, verbose=True)
        np.testing.assert_allclose(x, y)

    @pytest.mark.parametrize('dtype', [np.float64, np.complex128])
    def test_multi_rhs_vector(self, dtype):
        M = np.array([
            [1, 0, 2],
            [0, 0, 3],
            [-4, 5, 6],
        ], dtype=dtype)
        sM = scipy.sparse.csr_matrix(M)
        N = np.array([
            [3, 0, 1],
            [0, 2, 0],
            [0, 0, 0],
        ], dtype=dtype)
        sX = mkl_spsolve(sM, N, verbose=True)
        X = scipy.linalg.solve(M, N)
        np.testing.assert_allclose(X, sX)

    def test_rhs_shape_is_maintained(self):
        A = scipy.sparse.csr_matrix(np.array([
            [1, 0, 2],
            [0, 0, 3],
            [-4, 5, 6],
        ], dtype=np.complex128))
        b = np.array([0, 2, 0], dtype=np.complex128)
        out = mkl_spsolve(A, b, verbose=True)
        assert b.shape == out.shape

        b = np.array([0, 2, 0], dtype=np.complex128).reshape((3, 1))
        out = mkl_spsolve(A, b, verbose=True)
        assert b.shape == out.shape

    def test_sparse_rhs(self):
        A = scipy.sparse.csr_matrix([
            [1, 2, 0],
            [0, 3, 0],
            [0, 0, 5],
        ])
        b = scipy.sparse.csr_matrix([
            [0, 1],
            [1, 0],
            [0, 0],
        ])
        x = mkl_spsolve(A, b, verbose=True)
        ans = np.array([[-0.66666667, 1],
                        [0.33333333, 0],
                        [0, 0]])
        np.testing.assert_allclose(x.toarray(), ans)

    @pytest.mark.parametrize('dtype', [np.float64, np.complex128])
    def test_symmetric_solver(self, dtype):
        A = qutip.rand_herm(np.arange(1, 11)).data
        if dtype == np.float64:
            A = A.real
        x = np.ones(10, dtype=dtype)
        b = A.dot(x)
        y = mkl_spsolve(A, b, hermitian=1, verbose=True)
        np.testing.assert_allclose(x, y)


class Test_splu:
    @pytest.mark.parametrize('dtype', [np.float64, np.complex128])
    def test_repeated_rhs_solve(self, dtype):
        M = np.array([
            [1, 0, 2],
            [0, 0, 3],
            [-4, 5, 6],
        ], dtype=dtype)
        sM = scipy.sparse.csr_matrix(M)
        N = np.array([
            [3, 0, 1],
            [0, 2, 0],
            [0, 0, 0],
        ], dtype=dtype)
        test_X = np.zeros((3, 3), dtype=dtype)
        lu = mkl_splu(sM, verbose=True)
        for k in range(3):
            test_X[:, k] = lu.solve(N[:, k])
        lu.delete()
        expected_X = scipy.linalg.solve(M, N)
        np.testing.assert_allclose(test_X, expected_X)



__all__ = ["krylovsolve"]
"""
This module provides approximations of the time evolution operator
using small dimensional Krylov subspaces.
"""

from scipy.optimize import root_scalar
from math import ceil
import numpy as np
import warnings

from qutip.expect import expect
from qutip.qobj import Qobj
from qutip.solver import Result, Options
from qutip.ui.progressbar import BaseProgressBar, TextProgressBar
from qutip.sparse import eigh


def krylovsolve(
    H: Qobj,
    psi0: Qobj,
    tlist: np.array,
    krylov_dim: int,
    e_ops=None,
    options=None,
    progress_bar: bool = None,
    sparse: bool = False,
):
    """
    Time evolution of state vectors for time independent Hamiltonians.
    Evolve the state vector ("psi0") finding an approximation for the time
    evolution operator of Hamiltonian ("H") by obtaining the projection of
    the time evolution operator on a set of small dimensional Krylov
    subspaces (m << dim(H)).

    The output is either the state vector or the expectation values of
    supplied operators ("e_ops") at arbitrary points at ("tlist").

    **Additional options**

    Additional options to krylovsolve can be set with the following:

    * "store_states": stores states even though expectation values are
      requested via the "e_ops" argument.

    * "store_final_state": store final state even though expectation values are
      requested via the "e_ops" argument.

    Parameters
    ----------
    H : :class:`qutip.Qobj`
        System Hamiltonian.
    psi0 : :class: `qutip.Qobj`
        Initial state vector (ket).
    tlist : None / *list* / *array*
        List of times on which to evolve the initial state. If None, nothing
        happens but the code won't break.
    krylov_dim: int
        Dimension of Krylov approximation subspaces used for the time
        evolution approximation.
    e_ops : None / list of :class:`qutip.Qobj`
        Single operator or list of operators for which to evaluate
        expectation values.
    options : Options
        Instance of ODE solver options, as well as krylov parameters.
            atol: controls (approximately) the error desired for the final
                  solution. (Defaults to 1e-8)
            nsteps: maximum number of krylov's internal number of Lanczos
                    iterations. (Defaults to 10000)
    progress_bar : None / BaseProgressBar
         Optional instance of BaseProgressBar, or a subclass thereof, for
         showing the progress of the simulation.
    sparse : bool (default False)
         Use np.array to represent system Hamiltonians. If True, scipy sparse
         arrays are used instead.

    Returns
    -------
    result: :class:`qutip.Result`
        An instance of the class :class:`qutip.Result`, which contains
        either an *array* `result.expect` of expectation values for the times
        `tlist`, or an *array* `result.states` of state vectors corresponding
        to the times `tlist` [if `e_ops` is an empty list].
    """
    # check the physics
    _check_inputs(H, psi0, krylov_dim)

    # check extra inputs
    e_ops, e_ops_dict = _check_e_ops(e_ops)
    pbar = _check_progress_bar(progress_bar)

    # transform inputs type from Qobj to np.ndarray/csr_matrix
    if sparse:
        _H = H.get_data()  # (fast_) csr_matrix
    else:
        _H = H.full().copy()  # np.ndarray
    _psi = psi0.full().copy()
    _psi = _psi / np.linalg.norm(_psi)

    # create internal variable and output containers
    if options is None:
        options = Options(nsteps=10000)
    krylov_results = Result()
    krylov_results.solver = "krylovsolve"

    # handle particular cases of an empty tlist or single element
    n_tlist_steps = len(tlist)
    if n_tlist_steps < 1:
        return krylov_results

    if n_tlist_steps == 1:  # if tlist has only one element, return it
        krylov_results = particular_tlist_or_happy_breakdown(
            tlist, n_tlist_steps, options, psi0, e_ops, krylov_results, pbar
        )  # this will also raise a warning
        return krylov_results

    tf = tlist[-1]
    t0 = tlist[0]

    # optimization step using Lanczos, then reuse it for the first partition
    dim_m = krylov_dim
    krylov_basis, T_m = lanczos_algorithm(
        _H, _psi, krylov_dim=dim_m, sparse=sparse
    )

    # check if a happy breakdown occurred
    if T_m.shape[0] < krylov_dim + 1:
        if T_m.shape[0] == 1:
            # this means that the state does not evolve in time, it lies in a
            # symmetry of H subspace. Thus, theres no work to be done.
            krylov_results = particular_tlist_or_happy_breakdown(
                tlist,
                n_tlist_steps,
                options,
                psi0,
                e_ops,
                krylov_results,
                pbar,
                happy_breakdown=True,
            )
            return krylov_results
        else:
            # no optimization is required, convergence is guaranteed.
            delta_t = tf - t0
            n_timesteps = 1
    else:

        # calculate optimal number of internal timesteps.
        delta_t = _optimize_lanczos_timestep_size(
            T_m, krylov_basis=krylov_basis, tlist=tlist, options=options
        )
        n_timesteps = int(ceil((tf - t0) / delta_t))

        if n_timesteps >= options.nsteps:
            raise Exception(
                f"Optimization requires a number {n_timesteps} of lanczos iterations, "
                f"which exceeds the defined allowed number {options.nsteps}. This can "
                "be increased via the 'Options.nsteps' property."
            )

    partitions = _make_partitions(tlist=tlist, n_timesteps=n_timesteps)

    if progress_bar:
        pbar.start(len(partitions))

    # update parameters regarding e_ops
    krylov_results, expt_callback, options, n_expt_op = _e_ops_outputs(
        krylov_results, e_ops, n_tlist_steps, options
    )

    # parameters for the lazy iteration evolve tlist
    psi_norm = np.linalg.norm(_psi)
    last_t = t0

    for idx, partition in enumerate(partitions):

        evolved_states = _evolve_krylov_tlist(
            H=_H,
            psi0=_psi,
            krylov_dim=dim_m,
            tlist=partition,
            t0=last_t,
            psi_norm=psi_norm,
            krylov_basis=krylov_basis,
            T_m=T_m,
            sparse=sparse,
        )

        if idx == 0:
            krylov_basis = None
            T_m = None
            t_idx = 0

        _psi = evolved_states[-1]
        psi_norm = np.linalg.norm(_psi)
        last_t = partition[-1]

        # apply qobj to each evolved state, remove repeated tail elements
        qobj_evolved_states = [
            Qobj(state, dims=psi0.dims) for state in evolved_states[1:-1]
        ]

        krylov_results = _expectation_values(
            e_ops,
            n_expt_op,
            expt_callback,
            krylov_results,
            qobj_evolved_states,
            partitions,
            idx,
            t_idx,
            options,
        )

        t_idx += len(partition[1:-1])

        pbar.update(idx)

    pbar.finished()

    if e_ops_dict:
        krylov_results.expect = {
            e: krylov_results.expect[n]
            for n, e in enumerate(e_ops_dict.keys())
        }

    return krylov_results


def _expectation_values(
    e_ops,
    n_expt_op,
    expt_callback,
    res,
    evolved_states,
    partitions,
    idx,
    t_idx,
    options,
):

    if options.store_states:
        res.states += evolved_states

    for t, state in zip(
        range(t_idx, t_idx + len(partitions[idx][1:-1])), evolved_states
    ):

        if expt_callback:
            # use callback method
            res.expect.append(e_ops(t, state))

        for m in range(n_expt_op):
            op = e_ops[m]
            if not isinstance(op, Qobj) and callable(op):
                res.expect[m][t] = op(t, state)
                continue

            res.expect[m][t] = expect(op, state)

    if (
        idx == len(partitions) - 1
        and options.store_final_state
        and not options.store_states
    ):
        res.states = [evolved_states[-1]]

    return res


def lanczos_algorithm(
    H,
    psi: np.ndarray,
    krylov_dim: int,
    sparse: bool = False,
):
    """
    Computes a basis of the Krylov subspace for Hamiltonian 'H', a system
    state 'psi' and Krylov dimension 'krylov_dim'. The space is spanned
    by {psi, H psi, H^2 psi, ..., H^(krylov_dim) psi}.
    Parameters
    ------------
    H : np.ndarray or csr_matrix
       System Hamiltonian. If the Hamiltonian is dense, a np.ndarray is
       preferred, whereas if it is sparse, a scipy csr_matrix is optimal.
    psi: np.ndarray
        State used to calculate Krylov subspace.
    krylov_dim: int
        Dimension (krylov_dim + 1) of the spanned Krylov subspace.
    sparse: bool (optional, default False)
        Wether to perform scipy sparse matrix multiplication operations or
        numpy dense matrix multiplications.
    Returns
    ---------
    v: np.ndarray
        Lanczos eigenvector.
    T: np.ndarray
        Tridiagonal decomposition.
    """

    v = np.zeros((krylov_dim + 1, psi.shape[0]), dtype=complex)
    T_m = np.zeros((krylov_dim + 1, krylov_dim + 1), dtype=complex)

    v[0, :] = psi.squeeze()

    w_prime = H.dot(v[0, :])

    alpha = np.vdot(w_prime, v[0, :])

    w = w_prime - alpha * v[0, :]

    T_m[0, 0] = alpha

    for j in range(1, krylov_dim + 1):

        beta = np.linalg.norm(w)

        if beta < 1e-7:

            # Happy breakdown
            v_happy = v[0:j, :]
            T_m_happy = T_m[0:j, 0:j]

            return v_happy, T_m_happy

        v[j, :] = w / beta
        w_prime = H.dot(v[j, :])
        alpha = np.vdot(w_prime, v[j, :])

        w = w_prime - alpha * v[j, :] - beta * v[j - 1, :]

        T_m[j, j] = alpha
        T_m[j, j - 1] = beta
        T_m[j - 1, j] = beta

    return v, T_m


def _evolve(t0: float, krylov_basis: np.ndarray, T_m: np.ndarray):
    """
    Computes the time evolution operator 'U(t - t0) psi0_k', where 'psi0_k'
    is the first basis element of the Krylov subspace, as a function of time.
    Parameters
    ------------
    t0: float
        Initial time for the time evolution.
    krylov_basis: np.ndarray
        Krylov basis projector operator.
    T_m: np.ndarray
        Tridiagonal matrix decomposition of the system given by lanczos
        algorithm.
    Returns
    ---------
    time_evolution: function
        Time evolution given by the Krylov subspace approximation.
    """

    eigenvalues, eigenvectors = eigh(T_m)
    U = np.matmul(krylov_basis.T, eigenvectors)
    e0 = eigenvectors.conj().T[:, 0]

    def time_evolution(t):
        delta_t = t - t0
        aux = np.multiply(np.exp(-1j * delta_t * eigenvalues), e0)
        return np.matmul(U, aux)

    return time_evolution


def _evolve_krylov_tlist(
    H,
    psi0: np.ndarray,
    krylov_dim: int,
    tlist: list,
    t0: float,
    psi_norm: float = None,
    krylov_basis: np.array = None,
    T_m: np.array = None,
    sparse: bool = False,
):
    """
    Computes the Krylov approximation time evolution of dimension 'krylov_dim'
    for Hamiltonian 'H' and initial state 'psi0' for each time in 'tlist'.
    Parameters
    ------------
    H: np.ndarray or csr_matrix
        System Hamiltonian.
    psi0: np.ndarray
        Initial state vector.
    krylov_basis: np.ndarray
        Krylov basis projector operator.
    tlist: list
        List of timesteps for the time evolution.
    t0: float
        Initial time for the time evolution.
    psi_norm: float (optional, default False)
        Norm-2 of psi0.
    krylov_basis: np.ndarray (optional, default None)
        Krylov basis projector operator. If 'krylov_basis' is None, perform
        a lanczos iteration.
    T_m: np.ndarray (optional, default None)
        Tridiagonal matrix decomposition of the system given by lanczos
        algorithm. If 'T_m' is None, perform a lanczos iteration.
    Returns
    ---------
    psi_list: List[np.ndarray]
        List of evolved states at times t in 'tlist'.
    """

    if psi_norm is None:
        psi_norm = np.linalg.norm(psi0)

    if psi_norm != 1:
        psi = psi0 / psi_norm
    else:
        psi = psi0

    if (krylov_basis is None) or (T_m is None):
        krylov_basis, T_m = lanczos_algorithm(
            H=H, psi=psi, krylov_dim=krylov_dim, sparse=sparse
        )

    evolve = _evolve(t0, krylov_basis, T_m)
    psi_list = list(map(evolve, tlist))

    return psi_list


# ----------------------------------------------------------------------
# Auxiliar functions


def _check_inputs(H, psi0, krylov_dim):
    """Check that the inputs 'H' and 'psi0' have the correct structures."""
    if not isinstance(H, Qobj):
        raise TypeError(
            "krylovsolve currently supports Hamiltonian Qobj operators only"
        )

    if not H.isherm:
        raise TypeError("Hamiltonian 'H' must be hermician.")

    if not isinstance(psi0, Qobj):
        raise TypeError("'psi0' must be a Qobj.")

    if not psi0.isket:
        raise TypeError("Initial state must be a ket Qobj.")

    if not ((len(H.shape) == 2) and (H.shape[0] == H.shape[1])):
        raise ValueError("the Hamiltonian must be 2-dimensional square Qobj.")

    if not (psi0.dims[0] == H.dims[0]):
        raise ValueError(
            "'psi0' and the Hamiltonian must share the same dimension."
        )

    if not (H.shape[0] >= krylov_dim):
        raise ValueError(
            "the Hamiltonian dimension must be greater or equal to the maximum"
            " allowed krylov dimension 'krylov_dim'."
        )


def _check_e_ops(e_ops):
    """
    Check instances of e_ops and return the formatted version of e_ops
    and e_ops_dict.
    """
    if e_ops is None:
        e_ops = []
    if isinstance(e_ops, Qobj):
        e_ops = [e_ops]
    if isinstance(e_ops, dict):
        e_ops_dict = e_ops
        e_ops = [e for e in e_ops.values()]
    else:
        e_ops_dict = None
    return e_ops, e_ops_dict


def _check_progress_bar(progress_bar):
    """
    Check instance of progress_bar and return the object.
    """
    if progress_bar is None:
        pbar = BaseProgressBar()
    if progress_bar is True:
        pbar = TextProgressBar()
    return pbar


def particular_tlist_or_happy_breakdown(
    tlist,
    n_tlist_steps,
    options,
    psi0,
    e_ops,
    res,
    progress_bar,
    happy_breakdown=False,
):
    """Deals with the problem when 'tlist' contains a single element, where
    that same ket is returned and evaluated at 'e_ops', if provided.
    """

    if len(tlist) == 0:
        warnings.warn(
            "Input 'tlist' contains a single element. If 'e_ops' were provided"
            ", return its corresponding expectation values at 'psi0', else "
            "return 'psi0'."
        )

    progress_bar.start(1)

    res, expt_callback, options, n_expt_op = _e_ops_outputs(
        res, e_ops, n_tlist_steps, options
    )

    if options.store_states:
        res.states = [psi0]

    e_0 = None
    if expt_callback:
        # use callback method
        e_0 = e_ops(0, psi0)
        res.expect.append(e_0)

    e_m_0 = []
    for m in range(n_expt_op):
        op = e_ops[m]

        if not isinstance(op, Qobj) and callable(op):
            e_m_0.append(op(0, psi0))
            res.expect[m][0] = e_m_0[m]
            continue

        e_m_0.append(expect(op, psi0))
        res.expect[m][0] = e_m_0[m]

    if happy_breakdown:
        res = _happy_breakdown(
            tlist,
            options,
            res,
            psi0,
            expt_callback,
            e_0,
            n_expt_op,
            e_ops,
            e_m_0,
        )

    if (options.store_final_state) and (not options.store_states):
        res.states = [psi0]

    progress_bar.update(1)
    progress_bar.finished()
    return res


def _happy_breakdown(
    tlist, options, res, psi0, expt_callback, e_0, n_expt_op, e_ops, e_m_0
):
    """
    Dummy evolves the system if a happy breakdown of an eigenstate occurs.
    """
    for i in range(1, len(tlist)):
        if options.store_states:
            res.states.append(psi0)
        if expt_callback:
            res.expect.append(e_0)

        for m in range(n_expt_op):
            op = e_ops[m]
            res.expect[m][i] = e_m_0[m]
    return res


def _optimize_lanczos_timestep_size(T, krylov_basis, tlist, options):
    """
    Solves the equation defined to optimize the number of Lanczos
    iterations to be performed inside Krylov's algorithm.
    """

    f = _lanczos_error_equation_to_optimize_delta_t(
        T,
        krylov_basis=krylov_basis,
        t0=tlist[0],
        tf=tlist[-1],
        target_tolerance=options.atol,
    )

    # To avoid the singularity at t0, we add a small epsilon value
    t_min = (tlist[-1] - tlist[0]) / options.nsteps + tlist[0]
    bracket = [t_min, tlist[-1]]

    if (np.sign(f(bracket[0])) == -1) and (np.sign(f(bracket[-1])) == -1):
        delta_t = tlist[-1] - tlist[0]
        return delta_t

    elif (np.sign(f(bracket[0])) == 1) and (np.sign(f(bracket[-1])) == 1):
        raise ValueError(
            "No solution exists with the given combination of parameters 'krylov_dim', "
            "tolerance = 'options.atol', maximum number allowed of krylov internal "
            "partitions = 'options.nsteps' and 'tlist'. Try reducing the tolerance, or "
            "increasing 'krylov_dim'. If nothing works, then a deeper analysis of the "
            "problem is recommended."
        )

    else:
        sol = root_scalar(f=f, bracket=bracket, method="brentq", xtol=options.atol)
        if sol.converged:
            delta_t = sol.root
            return delta_t
        else:
            raise Exception(
                "Method did not converge, try increasing 'krylov_dim', "
                "taking a lesser final time 'tlist[-1]' or decreasing the "
                "tolerance via Options().atol. "
                "If nothing works, this problem might not be suitable for "
                "Krylov or a deeper analysis might be required."
            )


def _lanczos_error_equation_to_optimize_delta_t(
    T, krylov_basis, t0, tf, target_tolerance
):
    """
    Function to optimize in order to obtain the optimal number of
    Lanczos algorithm iterations, governed by the optimal timestep size between
    Lanczos iteractions.
    """
    eigenvalues1, eigenvectors1 = eigh(T[0:, 0:])
    U1 = np.matmul(krylov_basis[0:, 0:].T, eigenvectors1)
    e01 = eigenvectors1.conj().T[:, 0]

    eigenvalues2, eigenvectors2 = eigh(T[0:-1, 0: T.shape[1] - 1])
    U2 = np.matmul(krylov_basis[0:-1, :].T, eigenvectors2)
    e02 = eigenvectors2.conj().T[:, 0]

    def f(t):
        delta_t = -1j * (t - t0)

        aux1 = np.multiply(np.exp(delta_t * eigenvalues1), e01)
        psi1 = np.matmul(U1, aux1)

        aux2 = np.multiply(np.exp(delta_t * eigenvalues2), e02)
        psi2 = np.matmul(U2, aux2)

        error = np.linalg.norm(psi1 - psi2)

        steps = max(1, (tf - t0) // (t - t0))
        return np.log10(error) + np.log10(steps) - np.log10(target_tolerance)

    return f


def _make_partitions(tlist, n_timesteps):
    """Generates an internal 'partitions' list of np.arrays to iterate Lanczos
    algorithms on each of them, based on 'tlist' and the optimized number of
    iterations 'n_timesteps'.
    """

    _tlist = np.copy(tlist)

    if n_timesteps == 1:
        _tlist = np.insert(_tlist, 0, tlist[0])
        _tlist = np.append(_tlist, tlist[-1])
        partitions = [_tlist]
        return partitions

    n_timesteps += 1
    krylov_tlist = np.linspace(tlist[0], tlist[-1], n_timesteps)
    krylov_partitions = [
        np.array(krylov_tlist[i: i + 2]) for i in range(n_timesteps - 1)
    ]
    partitions = []
    for krylov_partition in krylov_partitions:
        start = krylov_partition[0]
        end = krylov_partition[-1]
        condition = _tlist <= end
        partitions.append([start] + _tlist[condition].tolist() + [end])
        _tlist = _tlist[~condition]

    return partitions


def _e_ops_outputs(krylov_results, e_ops, n_tlist_steps, opt):
    krylov_results.expect = []
    if callable(e_ops):
        n_expt_op = 0
        expt_callback = True
        krylov_results.num_expect = 1
    elif isinstance(e_ops, list):
        n_expt_op = len(e_ops)
        expt_callback = False
        krylov_results.num_expect = n_expt_op
        if n_expt_op == 0:
            # fall back on storing states
            opt.store_states = True
        else:
            for op in e_ops:
                if not isinstance(op, Qobj) and callable(op):
                    krylov_results.expect.append(
                        np.zeros(n_tlist_steps, dtype=complex)
                    )
                    continue
                if op.isherm:
                    krylov_results.expect.append(np.zeros(n_tlist_steps))
                else:
                    krylov_results.expect.append(
                        np.zeros(n_tlist_steps, dtype=complex)
                    )

    else:
        raise TypeError("Expectation parameter must be a list or a function")

    return krylov_results, expt_callback, opt, n_expt_op



import numpy as np
from ctypes import POINTER, c_int, c_char, byref
from numpy.ctypeslib import ndpointer
import qutip.settings as qset
zcsrgemv = qset.mkl_lib.mkl_cspblas_zcsrgemv


def mkl_spmv(A, x):
    """
    sparse csr_spmv using MKL
    """
    m, _ = A.shape

    # Pointers to data of the matrix
    data = A.data.ctypes.data_as(ndpointer(np.complex128, ndim=1, flags='C'))
    indptr = A.indptr.ctypes.data_as(POINTER(c_int))
    indices = A.indices.ctypes.data_as(POINTER(c_int))

    # Allocate output, using same conventions as input
    if x.ndim == 1:
        y = np.empty(m, dtype=np.complex128, order='C')
    elif x.ndim == 2 and x.shape[1] == 1:
        y = np.empty((m, 1), dtype=np.complex128, order='C')
    else:
        raise Exception('Input vector must be 1D row or 2D column vector')

    # Now call MKL. This returns the answer in the last argument, which shares
    # memory with y.
    zcsrgemv(
        byref(c_char(bytes(b'N'))),
        byref(c_int(m)),
        data,
        indptr,
        indices,
        x.ctypes.data_as(ndpointer(np.complex128, ndim=1, flags='C')),
        y.ctypes.data_as(ndpointer(np.complex128, ndim=1, flags='C')),
    )
    return y



import numpy as np
from numpy.testing import (run_module_suite, assert_,
                        assert_equal, assert_almost_equal)
import scipy.sparse as sp

from qutip.fastsparse import fast_csr_matrix, fast_identity
from qutip.random_objects import (rand_dm, rand_herm,
                                  rand_ket, rand_unitary)
from qutip.cy.spmath import (zcsr_kron, zcsr_transpose, zcsr_adjoint,
                            zcsr_isherm)


def test_csr_kron():
    "spmath: zcsr_kron"
    num_test = 5
    for _ in range(num_test):
        ra = np.random.randint(2,100)
        rb = np.random.randint(2,100)
        A = rand_herm(ra,0.5).data
        B = rand_herm(rb,0.5).data
        As = A.tocsr(1)
        Bs = B.tocsr(1)
        C = sp.kron(As,Bs, format='csr')
        D = zcsr_kron(A, B)
        assert_almost_equal(C.data, D.data)
        assert_equal(C.indices, D.indices)
        assert_equal(C.indptr, D.indptr)

    for _ in range(num_test):
        ra = np.random.randint(2,100)
        rb = np.random.randint(2,100)
        A = rand_ket(ra,0.5).data
        B = rand_herm(rb,0.5).data
        As = A.tocsr(1)
        Bs = B.tocsr(1)
        C = sp.kron(As,Bs, format='csr')
        D = zcsr_kron(A, B)
        assert_almost_equal(C.data, D.data)
        assert_equal(C.indices, D.indices)
        assert_equal(C.indptr, D.indptr)

    for _ in range(num_test):
        ra = np.random.randint(2,100)
        rb = np.random.randint(2,100)
        A = rand_dm(ra,0.5).data
        B = rand_herm(rb,0.5).data
        As = A.tocsr(1)
        Bs = B.tocsr(1)
        C = sp.kron(As,Bs, format='csr')
        D = zcsr_kron(A, B)
        assert_almost_equal(C.data, D.data)
        assert_equal(C.indices, D.indices)
        assert_equal(C.indptr, D.indptr)

    for _ in range(num_test):
        ra = np.random.randint(2,100)
        rb = np.random.randint(2,100)
        A = rand_ket(ra,0.5).data
        B = rand_ket(rb,0.5).data
        As = A.tocsr(1)
        Bs = B.tocsr(1)
        C = sp.kron(As,Bs, format='csr')
        D = zcsr_kron(A, B)
        assert_almost_equal(C.data, D.data)
        assert_equal(C.indices, D.indices)
        assert_equal(C.indptr, D.indptr)


def test_zcsr_transpose():
    "spmath: zcsr_transpose"
    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_ket(ra,0.5).data
        B = A.T.tocsr()
        C = A.trans()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_herm(5,1.0/ra).data
        B = A.T.tocsr()
        C = A.trans()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_dm(ra,1.0/ra).data
        B = A.T.tocsr()
        C = A.trans()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_unitary(ra,1.0/ra).data
        B = A.T.tocsr()
        C = A.trans()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)


def test_zcsr_adjoint():
    "spmath: zcsr_adjoint"
    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_ket(ra,0.5).data
        B = A.conj().T.tocsr()
        C = A.adjoint()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_herm(5,1.0/ra).data
        B = A.conj().T.tocsr()
        C = A.adjoint()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_dm(ra,1.0/ra).data
        B = A.conj().T.tocsr()
        C = A.adjoint()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)

    for k in range(50):
        ra = np.random.randint(2,100)
        A = rand_unitary(ra,1.0/ra).data
        B = A.conj().T.tocsr()
        C = A.adjoint()
        x = np.all(B.data == C.data)
        y = np.all(B.indices == C.indices)
        z = np.all(B.indptr == C.indptr)
        assert_(x*y*z)


def test_zcsr_mult():
    "spmath: zcsr_mult"
    for k in range(50):
        A = rand_ket(10,0.5).data
        B = rand_herm(10,0.5).data

        C = A.tocsr(1)
        D = B.tocsr(1)

        ans1 = B*A
        ans2 = D*C
        ans2.sort_indices()
        x = np.all(ans1.data == ans2.data)
        y = np.all(ans1.indices == ans2.indices)
        z = np.all(ans1.indptr == ans2.indptr)
        assert_(x*y*z)

    for k in range(50):
        A = rand_ket(10,0.5).data
        B = rand_ket(10,0.5).dag().data

        C = A.tocsr(1)
        D = B.tocsr(1)

        ans1 = B*A
        ans2 = D*C
        ans2.sort_indices()
        x = np.all(ans1.data == ans2.data)
        y = np.all(ans1.indices == ans2.indices)
        z = np.all(ans1.indptr == ans2.indptr)
        assert_(x*y*z)

        ans1 = A*B
        ans2 = C*D
        ans2.sort_indices()
        x = np.all(ans1.data == ans2.data)
        y = np.all(ans1.indices == ans2.indices)
        z = np.all(ans1.indptr == ans2.indptr)
        assert_(x*y*z)

    for k in range(50):
        A = rand_dm(10,0.5).data
        B = rand_dm(10,0.5).data

        C = A.tocsr(1)
        D = B.tocsr(1)

        ans1 = B*A
        ans2 = D*C
        ans2.sort_indices()
        x = np.all(ans1.data == ans2.data)
        y = np.all(ans1.indices == ans2.indices)
        z = np.all(ans1.indptr == ans2.indptr)
        assert_(x*y*z)

    for k in range(50):
        A = rand_dm(10,0.5).data
        B = rand_herm(10,0.5).data

        C = A.tocsr(1)
        D = B.tocsr(1)

        ans1 = B*A
        ans2 = D*C
        ans2.sort_indices()
        x = np.all(ans1.data == ans2.data)
        y = np.all(ans1.indices == ans2.indices)
        z = np.all(ans1.indptr == ans2.indptr)
        assert_(x*y*z)


def test_zcsr_isherm():
    "spmath: zcsr_isherm"
    N = 100
    for kk in range(100):
        A = rand_herm(N, 0.1)
        B = rand_herm(N, 0.05) + 1j*rand_herm(N, 0.05)
        assert_(zcsr_isherm(A.data))
        assert_(zcsr_isherm(B.data)==0)


def test_zcsr_isherm_compare_implicit_zero():
    """
    Regression test for gh-1350, comparing explicitly stored values in the
    matrix (but below the tolerance for allowable Hermicity) to implicit zeros.
    """
    tol = 1e-12
    n = 10

    base = sp.csr_matrix(np.array([[1, tol * 1e-3j], [0, 1]]))
    base = fast_csr_matrix((base.data, base.indices, base.indptr), base.shape)
    # If this first line fails, the zero has been stored explicitly and so the
    # test is invalid.
    assert base.data.size == 3
    assert zcsr_isherm(base, tol=tol)
    assert zcsr_isherm(base.T, tol=tol)

    # A similar test if the structures are different, but it's not
    # Hermitian.
    base = sp.csr_matrix(np.array([[1, 1j], [0, 1]]))
    base = fast_csr_matrix((base.data, base.indices, base.indptr), base.shape)
    assert base.data.size == 3
    assert not zcsr_isherm(base, tol=tol)
    assert not zcsr_isherm(base.T, tol=tol)

    # Catch possible edge case where it shouldn't be Hermitian, but faulty loop
    # logic doesn't fully compare all rows.
    base = sp.csr_matrix(np.array([
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 0, 0],
    ], dtype=np.complex128))
    base = fast_csr_matrix((base.data, base.indices, base.indptr), base.shape)
    assert base.data.size == 1
    assert not zcsr_isherm(base, tol=tol)
    assert not zcsr_isherm(base.T, tol=tol)

    # Pure diagonal matrix.
    base = fast_identity(n)
    base.data *= np.random.rand(n)
    assert zcsr_isherm(base, tol=tol)
    assert not zcsr_isherm(base * 1j, tol=tol)

    # Larger matrices where all off-diagonal elements are below the absolute
    # tolerance, so everything should always appear Hermitian, but with random
    # patterns of non-zero elements.  It doesn't matter that it isn't Hermitian
    # if scaled up; everything is below absolute tolerance, so it should appear
    # so.  We also set the diagonal to be larger to the tolerance to ensure
    # isherm can't just compare everything to zero.
    for density in np.linspace(0.2, 1, 21):
        base = tol * 1e-2 * (np.random.rand(n, n) + 1j*np.random.rand(n, n))
        # Mask some values out to zero.
        base[np.random.rand(n, n) > density] = 0
        np.fill_diagonal(base, tol * 1000)
        nnz = np.count_nonzero(base)
        base = sp.csr_matrix(base)
        base = fast_csr_matrix((base.data, base.indices, base.indptr), (n, n))
        assert base.data.size == nnz
        assert zcsr_isherm(base, tol=tol)
        assert zcsr_isherm(base.T, tol=tol)

        # Similar test when it must be non-Hermitian.  We set the diagonal to
        # be real because we want to test off-diagonal implicit zeros, and
        # having an imaginary first element would automatically fail.
        nnz = 0
        while nnz <= n:
            # Ensure that we don't just have the real diagonal.
            base = tol * 1000j*np.random.rand(n, n)
            # Mask some values out to zero.
            base[np.random.rand(n, n) > density] = 0
            np.fill_diagonal(base, tol * 1000)
            nnz = np.count_nonzero(base)
        base = sp.csr_matrix(base)
        base = fast_csr_matrix((base.data, base.indices, base.indptr), (n, n))
        assert base.data.size == nnz
        assert not zcsr_isherm(base, tol=tol)
        assert not zcsr_isherm(base.T, tol=tol)


if __name__ == "__main__":
    run_module_suite()



#cython: language_level=3

from qutip.cy.sparse_structs cimport CSR_Matrix

cdef int _safe_multiply(int a, int b) except? -1

cdef void _zcsr_add(CSR_Matrix * A, CSR_Matrix * B,
                    CSR_Matrix * C, double complex alpha)

cdef int _zcsr_add_core(double complex * Adata, int * Aind, int * Aptr,
                        double complex * Bdata, int * Bind, int * Bptr,
                        double complex alpha,
                        CSR_Matrix * C,
                        int nrows, int ncols) nogil

cdef void _zcsr_mult(CSR_Matrix * A, CSR_Matrix * B, CSR_Matrix * C)


cdef void _zcsr_kron(CSR_Matrix * A, CSR_Matrix * B, CSR_Matrix * C) except *

cdef void _zcsr_kron_core(double complex * dataA, int * indsA, int * indptrA,
                          double complex * dataB, int * indsB, int * indptrB,
                          CSR_Matrix * out,
                          int rowsA, int rowsB, int colsB) nogil

cdef void _zcsr_transpose(CSR_Matrix * A, CSR_Matrix * B)

cdef void _zcsr_adjoint(CSR_Matrix * A, CSR_Matrix * B)



__all__ = ['berry_curvature', 'plot_berry_curvature']

from qutip import (Qobj, tensor, basis, qeye, isherm, sigmax, sigmay, sigmaz)
import numpy as np

try:
    import matplotlib.pyplot as plt
except:
    pass


def berry_curvature(eigfs):
    """Computes the discretized Berry curvature on the two dimensional grid
    of parameters. The function works well for cases with no band mixing.

    Parameters
    ==========
    eigfs : numpy ndarray
        4 dimensional numpy ndarray where the first two indices are for the two
        discrete values of the two parameters and the third is the index of the
        occupied bands. The fourth dimension holds the eigenfunctions.

    Returns
    -------
    b_curv : numpy ndarray
        A two dimensional array of the discretized Berry curvature defined for
        the values of the two parameters defined in the eigfs.
    """
    nparam0 = eigfs.shape[0]
    nparam1 = eigfs.shape[1]
    nocc = eigfs.shape[2]
    b_curv = np.zeros((nparam0-1, nparam1-1), dtype=float)

    for i in range(nparam0-1):
        for j in range(nparam1-1):
            rect_prd = np.identity(nocc, dtype=complex)
            innP0 = np.zeros([nocc, nocc], dtype=complex)
            innP1 = np.zeros([nocc, nocc], dtype=complex)
            innP2 = np.zeros([nocc, nocc], dtype=complex)
            innP3 = np.zeros([nocc, nocc], dtype=complex)

            for k in range(nocc):
                for l in range(nocc):
                    wf0 = eigfs[i, j, k, :]
                    wf1 = eigfs[i+1, j, l, :]
                    innP0[k, l] = np.dot(wf0.conjugate(), wf1)

                    wf1 = eigfs[i+1, j, k, :]
                    wf2 = eigfs[i+1, j+1, l, :]
                    innP1[k, l] = np.dot(wf1.conjugate(), wf2)

                    wf2 = eigfs[i+1, j+1, k, :]
                    wf3 = eigfs[i, j+1, l, :]
                    innP2[k, l] = np.dot(wf2.conjugate(), wf3)

                    wf3 = eigfs[i, j+1, k, :]
                    wf0 = eigfs[i, j, l, :]
                    innP3[k, l] = np.dot(wf3.conjugate(), wf0)

            rect_prd = np.dot(rect_prd, innP0)
            rect_prd = np.dot(rect_prd, innP1)
            rect_prd = np.dot(rect_prd, innP2)
            rect_prd = np.dot(rect_prd, innP3)

            dett = np.linalg.det(rect_prd)
            curl_z = np.angle(dett)
            b_curv[i, j] = curl_z

    return b_curv


def plot_berry_curvature(eigfs):
    """Plots the discretized Berry curvature on the two dimensional grid
    of parameters. The function works well for cases with no band mixing."""
    b_curv = berry_curvature(eigfs)
    fig, ax = plt.subplots()
    ax.imshow(b_curv, origin="lower")
    ax.set_title("Berry curvature")
    ax.set_xlabel(r"$Parameter0$")
    ax.set_ylabel(r"$Parameter1$")
    fig.tight_layout()
    fig.savefig("berry_curvature.pdf")



from qutip.about import about
from qutip import settings as qset

def run(full=False):
    """
    Run the test scripts for QuTiP.

    Parameters
    ----------
    full: bool
        If True run all test (30 min). Otherwise skip few variants of the
        slowest tests.
    """
    # Call about to get all version info printed with tests
    about()
    import pytest
    real_num_cpu = qset.num_cpus
    real_thresh = qset.openmp_thresh
    if qset.has_openmp:
        # For travis which VMs have only 1 cpu.
        # Make sure the openmp version of the functions are tested.
        qset.num_cpus = 2
        qset.openmp_thresh = 100

    test_options = ["--verbosity=1", "--disable-pytest-warnings", "--pyargs"]
    if not full:
        test_options += ['-m', 'not slow']
    pytest.main(test_options + ["qutip"])
    # runs tests in qutip.tests module only

    # Restore previous settings
    if qset.has_openmp:
        qset.num_cpus = real_num_cpu
        qset.openmp_thresh = real_thresh
